from collections.abc import Callable, Mapping
from enum import Enum, auto
from typing import assert_never

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import _symmetric_modulo
from ....utils.bindings import Parameter
from ..typing import F, Fi, Ns, Ps, np_sndarray
from .abc import AnyExpr, Expr
from .addsub import ScalarSubtract
from .constfold import ScalarFoldedConstant
from .divmul import ScalarDivide, ScalarMultiply
from .group import Group
from .literal import Number
from .neg import ScalarNegate


class ScalarSymmetricModulo(Expr[AnyExpr, AnyExpr]):
    __slots__: tuple[str, ...] = ("_a", "_b")
    _a: AnyExpr
    _b: AnyExpr

    def __init__(self, a: AnyExpr, b: AnyExpr) -> None:
        self._a = a
        self._b = b

    @property
    @override
    def args(self) -> tuple[AnyExpr, AnyExpr]:
        return (self._a, self._b)

    @override
    def with_args(self, a: AnyExpr, b: AnyExpr) -> "ScalarSymmetricModulo":
        return ScalarSymmetricModulo(a, b)

    @override
    def constant_fold(self, dtype: np.dtype[Fi]) -> Fi | AnyExpr:
        return ScalarFoldedConstant.constant_fold_binary(
            self._a,
            self._b,
            dtype,
            lambda a, b: _symmetric_modulo(a, b),
            ScalarSymmetricModulo,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return _symmetric_modulo(
            self._a.eval(Xs, late_bound), self._b.eval(Xs, late_bound)
        )

    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        assert False, "cannot compute the data bounds for symmetric_modulo"

    @override
    def __repr__(self) -> str:
        return f"symmetric_modulo({self._a!r}, {self._b!r})"


class FiniteDifference(Enum):
    """
    Different types of finite differences.
    """

    central = auto()
    r"""
    Central finite difference, computed over the indices
    $\{i-k; \ldots; i; \ldots; i+k\}$.
    """

    forward = auto()
    r"""
    Forward finite difference, computed over the indices $\{i; \ldots; i+k\}$.
    """

    backwards = auto()
    r"""
    Backward finite difference, computed over the indices $\{i-k; \ldots; i\}$.
    """


def finite_difference_offsets(
    type: FiniteDifference,
    order: int,
    accuracy: int,
) -> tuple[int, ...]:
    match type:
        case FiniteDifference.central:
            noffsets = order + (order % 2) - 1 + accuracy
            p = (noffsets - 1) // 2
            return (0, *tuple(j for i in range(1, p + 1) for j in (i, -i)))
        case FiniteDifference.forward:
            return tuple(i for i in range(order + accuracy))
        case FiniteDifference.backwards:
            return tuple(-i for i in range(order + accuracy))
        case _:
            assert_never(type)


def finite_difference_coefficients(
    order: int,
    offsets: tuple[AnyExpr, ...],
    centre_dist: Callable[[AnyExpr], AnyExpr] = lambda x: x,
    delta_transform: Callable[[AnyExpr], AnyExpr] = lambda x: x,
) -> tuple[AnyExpr, ...]:
    """
    Finite difference coefficient algorithm from:

    Fornberg, B. (1988). Generation of finite difference formulas on arbitrarily
    spaced grids. *Mathematics of Computation*, 51(184), 699-706. Available from:
    [doi:10.1090/s0025-5718-1988-0935077-0](https://doi.org/10.1090/s0025-5718-1988-0935077-0).
    """

    dx0 = centre_dist
    M: int = order
    a: tuple[AnyExpr, ...] = offsets
    N: int = len(a) - 1

    # we explicitly the coefficient fraction into numerator and denominator
    #  expressions to delay the division for as long as possible, which allows
    #  more symbolic integer constant folding to occur
    coeffs_num: dict[tuple[int, int, int], AnyExpr] = {
        (0, 0, 0): Number.ONE,
    }
    coeffs_denom: dict[tuple[int, int, int], AnyExpr] = {
        (0, 0, 0): Number.ONE,
    }

    c1: AnyExpr = Number.ONE

    for n in range(1, N + 1):
        c2: AnyExpr = Number.ONE
        for v in range(0, n):
            c3: AnyExpr = delta_transform(Group(ScalarSubtract(a[n], a[v])))
            c2 = Group(ScalarMultiply(c2, c3))
            if n <= M:
                coeffs_num[(n, n - 1, v)] = Number.ZERO
                coeffs_denom[(n, n - 1, v)] = Number.ONE
            for m in range(0, min(n, M) + 1):
                if m > 0:
                    coeffs_num[(m, n, v)] = Group(
                        ScalarSubtract(
                            Group(
                                ScalarMultiply(
                                    ScalarMultiply(
                                        delta_transform(dx0(a[n])),
                                        coeffs_num[(m, n - 1, v)],
                                    ),
                                    coeffs_denom[(m - 1, n - 1, v)],
                                )
                            ),
                            Group(
                                ScalarMultiply(
                                    ScalarMultiply(
                                        Number.from_symbolic_int(m),
                                        coeffs_num[(m - 1, n - 1, v)],
                                    ),
                                    coeffs_denom[(m, n - 1, v)],
                                ),
                            ),
                        )
                    )
                    coeffs_denom[(m, n, v)] = Group(
                        ScalarMultiply(
                            ScalarMultiply(
                                coeffs_denom[(m, n - 1, v)],
                                coeffs_denom[(m - 1, n - 1, v)],
                            ),
                            c3,
                        )
                    )
                else:
                    coeffs_num[(m, n, v)] = Group(
                        ScalarMultiply(
                            delta_transform(dx0(a[n])), coeffs_num[(m, n - 1, v)]
                        )
                    )
                    coeffs_denom[(m, n, v)] = Group(
                        ScalarMultiply(coeffs_denom[(m, n - 1, v)], c3)
                    )
        for m in range(0, min(n, M) + 1):
            if m > 0:
                coeffs_num[(m, n, n)] = Group(
                    ScalarMultiply(
                        c1,
                        Group(
                            ScalarSubtract(
                                Group(
                                    ScalarMultiply(
                                        ScalarMultiply(
                                            Number.from_symbolic_int(m),
                                            coeffs_num[(m - 1, n - 1, n - 1)],
                                        ),
                                        coeffs_denom[(m, n - 1, n - 1)],
                                    )
                                ),
                                Group(
                                    ScalarMultiply(
                                        ScalarMultiply(
                                            delta_transform(dx0(a[n - 1])),
                                            coeffs_num[(m, n - 1, n - 1)],
                                        ),
                                        coeffs_denom[(m - 1, n - 1, n - 1)],
                                    )
                                ),
                            )
                        ),
                    )
                )
                coeffs_denom[(m, n, n)] = Group(
                    ScalarMultiply(
                        ScalarMultiply(
                            coeffs_denom[(m - 1, n - 1, n - 1)],
                            coeffs_denom[(m, n - 1, n - 1)],
                        ),
                        c2,
                    )
                )
            else:
                coeffs_num[(m, n, n)] = Group(
                    ScalarMultiply(
                        ScalarNegate(c1),
                        ScalarMultiply(
                            delta_transform(dx0(a[n - 1])),
                            coeffs_num[(m, n - 1, n - 1)],
                        ),
                    )
                )
                coeffs_denom[(m, n, n)] = Group(
                    ScalarMultiply(c2, coeffs_denom[(m, n - 1, n - 1)])
                )
        c1 = c2

    return tuple(
        Group(ScalarDivide(coeffs_num[M, N, v], coeffs_denom[M, N, v]))
        for v in range(0, N + 1)
    )



__all__ = ["QoILexer"]

from contextlib import contextmanager

from sly import Lexer

from ...utils.error import ctx


class QoILexer(Lexer):
    # === token declarations ===
    tokens = {
        # literals
        INTEGER,  # type: ignore[name-defined]  # noqa: F821
        FLOAT,  # type: ignore[name-defined]  # noqa: F821
        INF,  # type: ignore[name-defined]  # noqa: F821
        NAN,  # type: ignore[name-defined]  # noqa: F821
        STRING,  # type: ignore[name-defined]  # noqa: F821
        # operators
        PLUS,  # type: ignore[name-defined]  # noqa: F821
        MINUS,  # type: ignore[name-defined]  # noqa: F821
        POWER,  # type: ignore[name-defined]  # noqa: F821
        TIMES,  # type: ignore[name-defined]  # noqa: F821
        DIVIDE,  # type: ignore[name-defined]  # noqa: F821
        LESS_EQUAL,  # type: ignore[name-defined]  # noqa: F821
        LESS,  # type: ignore[name-defined]  # noqa: F821
        EQUAL,  # type: ignore[name-defined]  # noqa: F821
        NOT_EQUAL,  # type: ignore[name-defined]  # noqa: F821
        GREATER_EQUAL,  # type: ignore[name-defined]  # noqa: F821
        GREATER,  # type: ignore[name-defined]  # noqa: F821
        ASSIGN,  # type: ignore[name-defined]  # noqa: F821
        # array transpose
        TRANSPOSE,  # type: ignore[name-defined]  # noqa: F821
        # groups
        LPAREN,  # type: ignore[name-defined]  # noqa: F821
        RPAREN,  # type: ignore[name-defined]  # noqa: F821
        LBRACK,  # type: ignore[name-defined]  # noqa: F821
        RBRACK,  # type: ignore[name-defined]  # noqa: F821
        # separators
        COMMA,  # type: ignore[name-defined]  # noqa: F821
        COLON,  # type: ignore[name-defined]  # noqa: F821
        SEMI,  # type: ignore[name-defined]  # noqa: F821
        # identifiers
        ID,  # type: ignore[name-defined]  # noqa: F821
        # statements
        RETURN,  # type: ignore[name-defined]  # noqa: F821
        # constants
        EULER,  # type: ignore[name-defined]  # noqa: F821
        PI,  # type: ignore[name-defined]  # noqa: F821
        # data, late-bound constants, variables
        XS,  # type: ignore[name-defined]  # noqa: F821
        XA,  # type: ignore[name-defined]  # noqa: F821
        CS,  # type: ignore[name-defined]  # noqa: F821
        CA,  # type: ignore[name-defined]  # noqa: F821
        VS,  # type: ignore[name-defined]  # noqa: F821
        VA,  # type: ignore[name-defined]  # noqa: F821
        # array indexing
        IDX,  # type: ignore[name-defined]  # noqa: F821
        # functions
        # logarithms and exponentials
        LN,  # type: ignore[name-defined]  # noqa: F821
        LOG2,  # type: ignore[name-defined]  # noqa: F821
        LOG10,  # type: ignore[name-defined]  # noqa: F821
        LOG,  # type: ignore[name-defined]  # noqa: F821
        EXP,  # type: ignore[name-defined]  # noqa: F821
        EXP2,  # type: ignore[name-defined]  # noqa: F821
        EXP10,  # type: ignore[name-defined]  # noqa: F821
        # exponentiation
        SQRT,  # type: ignore[name-defined]  # noqa: F821
        SQUARE,  # type: ignore[name-defined]  # noqa: F821
        RECIPROCAL,  # type: ignore[name-defined]  # noqa: F821
        # absolute value
        ABS,  # type: ignore[name-defined]  # noqa: F821
        # sign and rounding
        SIGN,  # type: ignore[name-defined]  # noqa: F821
        FLOOR,  # type: ignore[name-defined]  # noqa: F821
        CEIL,  # type: ignore[name-defined]  # noqa: F821
        TRUNC,  # type: ignore[name-defined]  # noqa: F821
        ROUND_TIES_EVEN,  # type: ignore[name-defined]  # noqa: F821
        # trigonometric
        SIN,  # type: ignore[name-defined]  # noqa: F821
        COS,  # type: ignore[name-defined]  # noqa: F821
        TAN,  # type: ignore[name-defined]  # noqa: F821
        ASIN,  # type: ignore[name-defined]  # noqa: F821
        ACOS,  # type: ignore[name-defined]  # noqa: F821
        ATAN,  # type: ignore[name-defined]  # noqa: F821
        # hyperbolic
        SINH,  # type: ignore[name-defined]  # noqa: F821
        COSH,  # type: ignore[name-defined]  # noqa: F821
        TANH,  # type: ignore[name-defined]  # noqa: F821
        ASINH,  # type: ignore[name-defined]  # noqa: F821
        ACOSH,  # type: ignore[name-defined]  # noqa: F821
        ATANH,  # type: ignore[name-defined]  # noqa: F821
        # classification
        ISFINITE,  # type: ignore[name-defined]  # noqa: F821
        ISINF,  # type: ignore[name-defined]  # noqa: F821
        ISNAN,  # type: ignore[name-defined]  # noqa: F821
        # combinators
        NOT,  # type: ignore[name-defined]  # noqa: F821
        ALL,  # type: ignore[name-defined]  # noqa: F821
        ANY,  # type: ignore[name-defined]  # noqa: F821
        WHERE,  # type: ignore[name-defined]  # noqa: F821
        # array operations
        SIZE,  # type: ignore[name-defined]  # noqa: F821
        SHAPE,  # type: ignore[name-defined]  # noqa: F821
        SUM,  # type: ignore[name-defined]  # noqa: F821
        MATMUL,  # type: ignore[name-defined]  # noqa: F821
        # finite difference
        FINITE_DIFFERENCE,  # type: ignore[name-defined]  # noqa: F821
        # keyword arguments
        #  logarithm
        BASE,  # type: ignore[name-defined]  # noqa: F821
        #  finite difference
        ORDER,  # type: ignore[name-defined]  # noqa: F821
        ACCURACY,  # type: ignore[name-defined]  # noqa: F821
        TYPE,  # type: ignore[name-defined]  # noqa: F821
        AXIS,  # type: ignore[name-defined]  # noqa: F821
        GRID_SPACING,  # type: ignore[name-defined]  # noqa: F821
        GRID_CENTRE,  # type: ignore[name-defined]  # noqa: F821
        GRID_PERIOD,  # type: ignore[name-defined]  # noqa: F821
    }

    # === ignored whitespace patterns ===
    ignore = " \t"
    ignore_comment = r"#[^\n]*"

    @_(r"\n+")  # type: ignore[name-defined]  # noqa: F821
    def ignore_newline(self, t):  # type: ignore[no-redef]  # noqa: F811
        self.lineno += t.value.count("\n")

    # === token definitions ===

    # literals
    @_(r"[0-9]+(\.[0-9]*)?(e(\+|-)?[0-9]*)?")  # type: ignore[name-defined]  # noqa: F821
    def FLOAT(self, t):
        if ("." in t.value) or ("e" in t.value):
            # floating-point literal
            self.assert_or_error(
                t.value[-1] in "0123456789",
                t,
                f"invalid floating-point literal `{t.value}` requires at least one digit after '.' and 'e'",
            )
            return t
        else:
            # integer literal, pre-parsed
            with self.with_error_context(t, "excessive integer literal"):
                t.value = int(t.value)
            t.type = "INTEGER"
            return t

    @_(r'"[^"]*["]?')  # type: ignore[name-defined]  # noqa: F821
    def STRING(self, t):
        self.assert_or_error(
            t.value[-1] == '"', t, 'invalid string literal with missing closing `"`'
        )
        t.value = t.value[1:-1]
        return t

    # operators
    PLUS = r"\+"
    MINUS = r"-"
    # comes before TIMES so as to not parse ** as TIMES TIMES
    POWER = r"\*\*"
    TIMES = r"\*"
    DIVIDE = r"/"
    # comes before LESS and ASSIGN so as to not parse <= as LESS ASSIGN
    LESS_EQUAL = r"<="
    LESS = r"<"
    # comes before ASSIGN so as to not parse == as ASSIGN ASSIGN
    EQUAL = r"=="
    # comes before ASSIGN so as to not parse != as ! ASSIGN
    NOT_EQUAL = r"!="
    # comes before GREATER and ASSIGN so as to not parse >= as GREATER ASSIGN
    GREATER_EQUAL = r">="
    GREATER = r">"
    ASSIGN = r"="

    # array transpose
    TRANSPOSE = r"\.T"

    # groups
    LPAREN = r"\("
    RPAREN = r"\)"
    LBRACK = r"\["
    RBRACK = r"\]"

    # separators
    COMMA = r","
    COLON = r":"
    SEMI = r";"

    # identifiers
    ID = r"[a-zA-Z_][a-zA-Z0-9_]*"

    # statements
    ID["return"] = RETURN  # type: ignore[index, name-defined]  # noqa: F821

    # literals
    ID["Inf"] = INF  # type: ignore[index, name-defined]  # noqa: F821
    ID["NaN"] = NAN  # type: ignore[index, name-defined]  # noqa: F821

    # constants
    ID["e"] = EULER  # type: ignore[index, name-defined]  # noqa: F821
    ID["pi"] = PI  # type: ignore[index, name-defined]  # noqa: F821

    # data, late-bound constants, variables
    ID["x"] = XS  # type: ignore[index, name-defined]  # noqa: F821
    ID["X"] = XA  # type: ignore[index, name-defined]  # noqa: F821
    ID["c"] = CS  # type: ignore[index, name-defined]  # noqa: F821
    ID["C"] = CA  # type: ignore[index, name-defined]  # noqa: F821
    ID["v"] = VS  # type: ignore[index, name-defined]  # noqa: F821
    ID["V"] = VA  # type: ignore[index, name-defined]  # noqa: F821

    # array indexing
    ID["I"] = IDX  # type: ignore[index, name-defined]  # noqa: F821

    # functions
    # logarithms and exponentials
    ID["ln"] = LN  # type: ignore[index, name-defined]  # noqa: F821
    ID["log2"] = LOG2  # type: ignore[index, name-defined]  # noqa: F821
    ID["log10"] = LOG10  # type: ignore[index, name-defined]  # noqa: F821
    ID["log"] = LOG  # type: ignore[index, name-defined]  # noqa: F821
    ID["exp"] = EXP  # type: ignore[index, name-defined]  # noqa: F821
    ID["exp2"] = EXP2  # type: ignore[index, name-defined]  # noqa: F821
    ID["exp10"] = EXP10  # type: ignore[index, name-defined]  # noqa: F821
    # exponentiation
    ID["sqrt"] = SQRT  # type: ignore[index, name-defined]  # noqa: F821
    ID["square"] = SQUARE  # type: ignore[index, name-defined]  # noqa: F821
    ID["reciprocal"] = RECIPROCAL  # type: ignore[index, name-defined]  # noqa: F821
    # absolute value
    ID["abs"] = ABS  # type: ignore[index, name-defined]  # noqa: F821
    # sign and rounding
    ID["sign"] = SIGN  # type: ignore[index, name-defined]  # noqa: F821
    ID["floor"] = FLOOR  # type: ignore[index, name-defined]  # noqa: F821
    ID["ceil"] = CEIL  # type: ignore[index, name-defined]  # noqa: F821
    ID["trunc"] = TRUNC  # type: ignore[index, name-defined]  # noqa: F821
    ID["round_ties_even"] = ROUND_TIES_EVEN  # type: ignore[index, name-defined]  # noqa: F821
    # trigonometric
    ID["sin"] = SIN  # type: ignore[index, name-defined]  # noqa: F821
    ID["cos"] = COS  # type: ignore[index, name-defined]  # noqa: F821
    ID["tan"] = TAN  # type: ignore[index, name-defined]  # noqa: F821
    ID["asin"] = ASIN  # type: ignore[index, name-defined]  # noqa: F821
    ID["acos"] = ACOS  # type: ignore[index, name-defined]  # noqa: F821
    ID["atan"] = ATAN  # type: ignore[index, name-defined]  # noqa: F821
    # hyperbolic
    ID["sinh"] = SINH  # type: ignore[index, name-defined]  # noqa: F821
    ID["cosh"] = COSH  # type: ignore[index, name-defined]  # noqa: F821
    ID["tanh"] = TANH  # type: ignore[index, name-defined]  # noqa: F821
    ID["asinh"] = ASINH  # type: ignore[index, name-defined]  # noqa: F821
    ID["acosh"] = ACOSH  # type: ignore[index, name-defined]  # noqa: F821
    ID["atanh"] = ATANH  # type: ignore[index, name-defined]  # noqa: F821
    # classification
    ID["isfinite"] = ISFINITE  # type: ignore[index, name-defined]  # noqa: F821
    ID["isinf"] = ISINF  # type: ignore[index, name-defined]  # noqa: F821
    ID["isnan"] = ISNAN  # type: ignore[index, name-defined]  # noqa: F821
    # combinators
    ID["not"] = NOT  # type: ignore[index, name-defined]  # noqa: F821
    ID["all"] = ALL  # type: ignore[index, name-defined]  # noqa: F821
    ID["any"] = ANY  # type: ignore[index, name-defined]  # noqa: F821
    ID["where"] = WHERE  # type: ignore[index, name-defined]  # noqa: F821
    # array operations
    ID["size"] = SIZE  # type: ignore[index, name-defined]  # noqa: F821
    ID["shape"] = SHAPE  # type: ignore[index, name-defined]  # noqa: F821
    ID["sum"] = SUM  # type: ignore[index, name-defined]  # noqa: F821
    ID["matmul"] = MATMUL  # type: ignore[index, name-defined]  # noqa: F821
    # finite difference
    ID["finite_difference"] = FINITE_DIFFERENCE  # type: ignore[index, name-defined]  # noqa: F821

    # keyword arguments
    #  logarithm
    ID["base"] = BASE  # type: ignore[index, name-defined]  # noqa: F821
    #  finite difference
    ID["order"] = ORDER  # type: ignore[index, name-defined]  # noqa: F821
    ID["accuracy"] = ACCURACY  # type: ignore[index, name-defined]  # noqa: F821
    ID["type"] = TYPE  # type: ignore[index, name-defined]  # noqa: F821
    ID["axis"] = AXIS  # type: ignore[index, name-defined]  # noqa: F821
    ID["grid_spacing"] = GRID_SPACING  # type: ignore[index, name-defined]  # noqa: F821
    ID["grid_centre"] = GRID_CENTRE  # type: ignore[index, name-defined]  # noqa: F821
    ID["grid_period"] = GRID_PERIOD  # type: ignore[index, name-defined]  # noqa: F821

    # === lexer error handling ===
    def error(self, t):
        self.raise_error(t, f"unexpected character `{t.value[0]}`")

    def raise_error(self, t, message):
        raise (
            SyntaxError(
                message,
                (
                    "<qoi>",
                    t.lineno,
                    self.find_column(t),
                    self.text.splitlines()[t.lineno - 1],
                ),
            )
            | ctx
        )

    def assert_or_error(self, check, t, message):
        if not check:
            self.raise_error(t, message)

    @contextmanager
    def with_error_context(self, t, message, exception=Exception):
        try:
            yield
        except exception:
            self.raise_error(t, message)

    def find_column(self, token):
        last_cr = self.text.rfind("\n", 0, token.index)
        if last_cr < 0:
            last_cr = 0
        column = (token.index - last_cr) + 1
        return column

    @staticmethod
    def token_to_name(token: str) -> str:
        return {
            # special
            "$end": "EOF",
            # literals
            "INTEGER": "integer",
            "FLOAT": "floating-point number",
            "INF": "`Inf`",
            "NAN": "`NaN`",
            "STRING": "string",
            # operators
            "PLUS": "`+`",
            "MINUS": "`-`",
            "POWER": "`**`",
            "TIMES": "`*`",
            "DIVIDE": "`/`",
            "LESS_EQUAL": "`<=`",
            "LESS": "`<`",
            "EQUAL": "`==`",
            "NOT_EQUAL": "`!=`",
            "GREATER_EQUAL": "`>=`",
            "GREATER": "`>`",
            "ASSIGN": "`=`",
            # array transpose
            "TRANSPOSE": "`.T`",
            # groups
            "LPAREN": "`(`",
            "RPAREN": "`)`",
            "LBRACK": "`[`",
            "RBRACK": "`]`",
            # separators
            "COMMA": "`,`",
            "COLON": "`:`",
            "SEMI": "`;`",
            # identifiers
            "ID": "identifier",
            # statements
            "RETURN": "`return`",
            # constants
            "EULER": "`e`",
            "PI": "`pi`",
            # data, late-bound constants, variables
            "XS": "`x`",
            "XA": "`X`",
            "CS": "`c`",
            "CA": "`C`",
            "VS": "`v`",
            "VA": "`V`",
            # array indexing
            "IDX": "`I`",
            # functions
            # logarithms and exponentials
            "LN": "`ln`",
            "LOG2": "`log2`",
            "LOG10": "`log10`",
            "LOG": "`log`",
            "EXP": "`exp`",
            "EXP2": "`exp2`",
            "EXP10": "`exp10`",
            # exponentiation
            "SQRT": "`sqrt`",
            "SQUARE": "`square`",
            "RECIPROCAL": "`reciprocal`",
            # absolute value
            "ABS": "`abs`",
            # sign and rounding
            "SIGN": "`sign`",
            "FLOOR": "`floor`",
            "CEIL": "`ceil`",
            "TRUNC": "`trunc`",
            "ROUND_TIES_EVEN": "`round_ties_even`",
            # trigonometric
            "SIN": "`sin`",
            "COS": "`cos`",
            "TAN": "`tan`",
            "ASIN": "`asin`",
            "ACOS": "`acos`",
            "ATAN": "`atan`",
            # hyperbolic
            "SINH": "`sinh`",
            "COSH": "`cosh`",
            "TANH": "`tanh`",
            "ASINH": "`asinh`",
            "ACOSH": "`acosh`",
            "ATANH": "`atanh`",
            # classification
            "ISFINITE": "`isfinite`",
            "ISINF": "`isinf`",
            "ISNAN": "`isnan`",
            # combinators
            "NOT": "`not`",
            "ALL": "`all`",
            "ANY": "`any`",
            "WHERE": "`where`",
            # array operations
            "SIZE": "`size`",
            "SHAPE": "`shape`",
            "SUM": "`sum`",
            "MATMUL": "`matmul`",
            # finite difference
            "FINITE_DIFFERENCE": "`finite_difference`",
            # keyword arguments
            #  logarithm
            "BASE": "`base`",
            #  finite difference
            "ORDER": "`order`",
            "ACCURACY": "`accuracy`",
            "TYPE": "`type`",
            "AXIS": "`axis`",
            "GRID_SPACING": "`grid_spacing`",
            "GRID_CENTRE": "`grid_centre`",
            "GRID_PERIOD": "`grid_period`",
        }.get(token, f"<{token}>")



"""
Error bound safeguard.
"""

__all__ = ["ErrorBoundSafeguard"]

from collections.abc import Set
from typing import ClassVar, Literal

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ...utils._compat import _ensure_array, _logical_and
from ...utils.bindings import Bindings, Parameter
from ...utils.cast import ToFloatMode, as_bits, saturating_finite_float_cast, to_float
from ...utils.error import TypeCheckError, ctx, lookup_enum_or_raise
from ...utils.intervals import Interval, IntervalUnion, Lower, Upper
from ...utils.typing import JSON, S, T
from ..eb import (
    ErrorBound,
    _apply_finite_error_bound,
    _check_error_bound,
    _compute_finite_absolute_error,
    _compute_finite_absolute_error_bound,
)
from .abc import PointwiseSafeguard


class ErrorBoundSafeguard(PointwiseSafeguard):
    """
    The `ErrorBoundSafeguard` guarantees that the pointwise error `type` is less than or equal to the provided bound `eb`.

    Infinite values are preserved with the same bit pattern. If `equal_nan` is
    set to [`True`][True], correcting a NaN value to a NaN value with a
    different bit pattern also satisfies the error bound. If `equal_nan` is set
    to [`False`][False], NaN values are also preserved with the same bit
    pattern.

    The error bound can be verified by casting the data and error bound to a
    sufficiently large floating-point type, selected by
    [`ToFloatMode.lossless.floating_point_dtype_for`][.....utils.cast.ToFloatMode.floating_point_dtype_for],
    using [`to_float`][.....utils.cast.to_float].

    Parameters
    ----------
    type : str | ErrorBound
        The type of error bound that is enforced by this safeguard.
    eb : int | float | str | Parameter
        The value of or late-bound parameter name for the error bound that is
        enforced by this safeguard.
    equal_nan: bool
        Whether correcting a NaN value to a NaN value with a different bit
        pattern satisfies the error bound.

    Raises
    ------
    TypeCheckError
        if any parameter has the wrong type.
    ValueError
        if `type` does not name a valid error bound.
    ValueError
        if `eb` is an invalid error bound value for the error bound `type`.
    """

    __slots__: tuple[str, ...] = ("_type", "_eb", "_equal_nan")
    _type: ErrorBound
    _eb: int | float | Parameter
    _equal_nan: bool

    kind: ClassVar[str] = "eb"

    def __init__(
        self,
        type: str | ErrorBound,
        eb: int | float | str | Parameter,
        *,
        equal_nan: bool = False,
    ) -> None:
        with ctx.safeguard(self):
            with ctx.parameter("type"):
                TypeCheckError.check_instance_or_raise(type, str | ErrorBound)
                self._type = (
                    type
                    if isinstance(type, ErrorBound)
                    else lookup_enum_or_raise(ErrorBound, type)
                )

            with ctx.parameter("eb"):
                TypeCheckError.check_instance_or_raise(
                    eb, int | float | str | Parameter
                )
                if isinstance(eb, Parameter):
                    self._eb = eb
                elif isinstance(eb, str):
                    self._eb = Parameter(eb)
                else:
                    _check_error_bound(self._type, eb)
                    self._eb = eb

            with ctx.parameter("equal_nan"):
                TypeCheckError.check_instance_or_raise(equal_nan, bool)
                self._equal_nan = equal_nan

    @property
    @override
    def late_bound(self) -> Set[Parameter]:
        """
        The set of late-bound parameters that this safeguard has.

        Late-bound parameters are only bound when checking and applying the
        safeguard, in contrast to the normal early-bound parameters that are
        configured during safeguard initialisation.

        Late-bound parameters can be used for parameters that depend on the
        specific data that is to be safeguarded.
        """

        return frozenset([self._eb]) if isinstance(self._eb, Parameter) else frozenset()

    @np.errstate(divide="ignore", over="ignore", under="ignore", invalid="ignore")
    @override
    def check_pointwise(
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Check which elements in the `prediction` array satisfy the error bound.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Original data array, relative to which the `prediction` is checked.
        prediction : np.ndarray[S, np.dtype[T]]
            Prediction for the `data` array.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only check at data points where the condition is [`True`][True].

        Returns
        -------
        ok : np.ndarray[S, np.dtype[np.bool]]
            Pointwise, `True` if the check succeeded for this element.

        Raises
        ------
        LateBoundParameterResolutionError
            if the error bound `eb` is late-bound but its late-bound parameter
            is not in `late_bound`.
        ValueError
            if the late-bound `eb` could not be broadcast to the `data`'s shape.
        ValueError
            if the late-bound `eb` is non-finite, i.e. infinite or NaN, or an
            invalid error bound value for the error bound `type`.
        """

        ftype: np.dtype[np.floating] = ToFloatMode.lossless.floating_point_dtype_for(
            data.dtype
        )
        data_float: np.ndarray[S, np.dtype[np.floating]] = to_float(data, ftype=ftype)
        prediction_float: np.ndarray[S, np.dtype[np.floating]] = to_float(
            prediction, ftype=ftype
        )

        with ctx.safeguard(self), ctx.parameter("eb"):
            eb: np.ndarray[tuple[()] | S, np.dtype[np.floating]] = (
                late_bound.resolve_ndarray_with_saturating_finite_float_cast(
                    self._eb,
                    data.shape,
                    ftype,
                )
                if isinstance(self._eb, Parameter)
                else saturating_finite_float_cast(self._eb, ftype)
            )
            if isinstance(self._eb, Parameter):
                _eb: Parameter = self._eb
                with ctx.late_bound_parameter(_eb):
                    _check_error_bound(self._type, eb)

        finite_ok: np.ndarray[S, np.dtype[np.bool]] = np.less_equal(
            _compute_finite_absolute_error(self._type, data_float, prediction_float),
            _compute_finite_absolute_error_bound(self._type, eb, data_float),
        )

        # bitwise equality for inf and NaNs (unless equal_nan)
        same_bits = as_bits(data) == as_bits(prediction)
        both_nan = self._equal_nan and (np.isnan(data) & np.isnan(prediction))

        ok: np.ndarray[S, np.dtype[np.bool]] = _ensure_array(finite_ok)
        np.copyto(ok, same_bits, where=~np.isfinite(data), casting="no")
        if self._equal_nan:
            np.copyto(ok, both_nan, where=np.isnan(data), casting="no")
        if where is not True:
            ok[~where] = True
        return ok

    @override
    def compute_safe_intervals(
        self,
        data: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> IntervalUnion[T, int, int]:
        """
        Compute the intervals in which the error bound is upheld with respect
        to the `data`.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Data for which the safe intervals should be computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the safe intervals at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        intervals : IntervalUnion[T, int, int]
            Union of intervals in which the error bound is upheld.

        Raises
        ------
        LateBoundParameterResolutionError
            if the error bound `eb` is late-bound but its late-bound parameter
            is not in `late_bound`.
        ValueError
            if the late-bound `eb` could not be broadcast to the `data`'s shape.
        ValueError
            if the late-bound `eb` is non-finite, i.e. infinite or NaN, or an
            invalid error bound value for the error bound `type`.
        """

        ftype: np.dtype[np.floating] = ToFloatMode.lossless.floating_point_dtype_for(
            data.dtype
        )

        with ctx.safeguard(self), ctx.parameter("eb"):
            eb: np.ndarray[tuple[()] | S, np.dtype[np.floating]] = (
                late_bound.resolve_ndarray_with_saturating_finite_float_cast(
                    self._eb,
                    data.shape,
                    ftype,
                )
                if isinstance(self._eb, Parameter)
                else saturating_finite_float_cast(self._eb, ftype)
            )
            if isinstance(self._eb, Parameter):
                _eb: Parameter = self._eb
                with ctx.late_bound_parameter(_eb):
                    _check_error_bound(self._type, eb)

        lower, upper = _apply_finite_error_bound(
            self._type, eb, data, to_float(data, ftype=ftype)
        )

        dataf: np.ndarray[tuple[int], np.dtype[T]] = data.flatten()
        valid = Interval.empty_like(dataf).preserve_inf(dataf)

        Lower(lower.flatten()) <= valid[np.isfinite(dataf)] <= Upper(upper.flatten())

        wheref: Literal[True] | np.ndarray[tuple[int], np.dtype[np.bool]] = (
            True if where is True else where.flatten()
        )

        return valid.preserve_any_nan(
            dataf, equal_nan=self._equal_nan
        ).preserve_only_where(wheref)

    @override
    def compute_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Compute the footprint of the `foot` array, e.g. for expanding data
        points into the pointwise checks that they contribute to.

        The footprint is equivalent to `foot & where`.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the footprint at pointwise checks where the condition
            is [`True`][True].

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The footprint of the `foot` array.
        """

        return _logical_and(foot, where)

    @override
    def compute_inverse_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Compute the inverse footprint of the `foot` array, e.g. for expanding
        pointwise check fails into the points that could have contributed to
        the failures.

        The inverse footprint is equivalent to `foot & where`.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the inverse footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the inverse footprint at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The inverse footprint of the `foot` array.
        """

        return _logical_and(foot, where)

    @override
    def get_config(self) -> dict[str, JSON]:
        """
        Returns the configuration of the safeguard.

        Returns
        -------
        config : dict[str, JSON]
            Configuration of the safeguard.
        """

        return dict(
            kind=type(self).kind,
            type=self._type.name,
            eb=str(self._eb) if isinstance(self._eb, Parameter) else self._eb,
            equal_nan=self._equal_nan,
        )



import sys
from collections.abc import Callable, Mapping
from warnings import warn

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import _broadcast_to, _floating_e, _floating_pi
from ....utils.bindings import Parameter
from ..typing import F, Ns, Ps, np_sndarray
from .abc import AnyExpr, EmptyExpr


class Number(EmptyExpr):
    __slots__: tuple[str, ...] = ("_n",)
    _n: str

    ZERO: "Number"
    ONE: "Number"
    TWO: "Number"

    def __init__(self, n: str) -> None:
        self._n = n

    @staticmethod
    def from_symbolic_int(n: int) -> "Number":
        try:
            return Number(f"{n}")
        except ValueError:
            pass

        warn(
            "symbolic integer evaluation, likely a**b, resulted in excessive "
            + "number of digits, try making one operand a floating-point "
            + "literal",
            category=RuntimeWarning,
        )

        int_max_str_digits = sys.get_int_max_str_digits()
        sys.set_int_max_str_digits(0)

        try:
            return Number(f"{n}")
        finally:
            sys.set_int_max_str_digits(int_max_str_digits)

    @staticmethod
    def from_symbolic_int_as_float(n: int, force_negative: bool = False) -> "Number":
        expr = Number.from_symbolic_int(n)
        if force_negative and not expr._n.startswith("-"):
            expr._n = f"-{expr._n}"
        expr._n = f"{expr._n}.0"
        return expr

    @property
    @override
    def args(self) -> tuple[()]:
        return ()

    @override
    def with_args(self) -> "Number":
        return Number(self._n)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return dtype.type(self._n)

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        n: F = Xs.dtype.type(self._n)
        return _broadcast_to(n, Xs.shape[:1])

    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        assert False, "number literals have no data bounds"

    def as_int(self) -> None | int:
        if (
            ("." in self._n)
            or ("e" in self._n)
            or ("inf" in self._n)
            or ("nan" in self._n)
        ):
            return None
        try:
            return int(self._n)
        except ValueError:
            pass

        warn(
            "symbolic integer evaluation, likely a**b, resulted in excessive "
            + "number of digits, try making one operand a floating-point "
            + "literal",
            category=RuntimeWarning,
        )

        int_max_str_digits = sys.get_int_max_str_digits()
        sys.set_int_max_str_digits(0)

        try:
            return int(self._n)
        finally:
            sys.set_int_max_str_digits(int_max_str_digits)

    @override
    def __repr__(self) -> str:
        return self._n

    @staticmethod
    def symbolic_fold_unary(expr: AnyExpr, m: Callable[[int], int]) -> "None | Number":
        if not isinstance(expr, Number):
            return None
        i = expr.as_int()
        if i is None:
            return None
        return Number.from_symbolic_int(m(i))

    @staticmethod
    def symbolic_fold_binary(
        a: AnyExpr, b: AnyExpr, m: Callable[[int, int], int]
    ) -> "None | Number":
        if not isinstance(a, Number) or not isinstance(b, Number):
            return None
        ai = a.as_int()
        bi = b.as_int()
        if (ai is None) or (bi is None):
            return None
        return Number.from_symbolic_int(m(ai, bi))


Number.ZERO = Number.from_symbolic_int(0)
Number.ONE = Number.from_symbolic_int(1)
Number.TWO = Number.from_symbolic_int(2)


class Pi(EmptyExpr):
    __slots__: tuple[str, ...] = ()

    @property
    @override
    def args(self) -> tuple[()]:
        return ()

    @override
    def with_args(self) -> "Pi":
        return Pi()

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return _floating_pi(dtype)

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        pi: F = _floating_pi(Xs.dtype)
        return _broadcast_to(pi, Xs.shape[:1])

    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        assert False, "pi has no data bounds"

    @override
    def __repr__(self) -> str:
        return "pi"


class Euler(EmptyExpr):
    __slots__: tuple[str, ...] = ()

    @property
    @override
    def args(self) -> tuple[()]:
        return ()

    @override
    def with_args(self) -> "Euler":
        return Euler()

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return _floating_e(dtype)

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        e: F = _floating_e(Xs.dtype)
        return _broadcast_to(e, Xs.shape[:1])

    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        assert False, "Euler's e has no data bounds"

    @override
    def __repr__(self) -> str:
        return "e"



from collections.abc import Mapping
from warnings import warn

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import _broadcast_to, _ensure_array, _is_of_shape, _ones
from ....utils.bindings import Parameter
from ....utils.error import QuantityOfInterestRuntimeWarning
from ..bound import DataBounds, data_bounds
from ..typing import F, Ns, Ps, np_sndarray
from .abc import AnyExpr, EmptyExpr


class Data(EmptyExpr):
    __slots__: tuple[str, ...] = ("_index",)
    _index: tuple[int, ...]

    SCALAR: "Data"

    def __init__(self, index: tuple[int, ...]) -> None:
        self._index = index

    @property
    def index(self) -> tuple[int, ...]:
        return self._index

    @property
    @override
    def args(self) -> tuple[()]:
        return ()

    @override
    def with_args(self) -> "Data":
        return Data(self._index)

    @property  # type: ignore
    @override
    def has_data(self) -> bool:
        return True

    @override  # type: ignore
    def eval_has_data(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[np.bool]]:
        # we could just return all True here, since all data is ... data
        # but we also currently guarantee that NaN data values stay NaN,
        #  i.e. that the values are constant, so not really data dependent
        # therefore, we say that only non-NaN data is actually data
        data = self.eval(Xs, late_bound)
        data_is_not_nan = _ensure_array(np.isnan(data))
        np.logical_not(data_is_not_nan, out=data_is_not_nan)
        return data_is_not_nan

    @property  # type: ignore
    @override
    def data_indices(self) -> frozenset[tuple[int, ...]]:
        return frozenset([self._index])

    @override  # type: ignore
    def apply_array_element_offset(
        self,
        axis: int,
        offset: int,
    ) -> "Data":
        index = list(self._index)
        index[axis] += offset
        return Data(index=tuple(index))

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return self

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        out: np.ndarray[tuple[int, ...], np.dtype[F]] = Xs[(...,) + self._index]
        assert _is_of_shape(out, Xs.shape[:1])
        return out

    @data_bounds(DataBounds.infallible)
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        exprv = Xs[(...,) + self._index]

        if not np.all((expr_lower <= exprv) | np.isnan(exprv)):
            warn(
                "data lower bounds are above the data values",
                category=QuantityOfInterestRuntimeWarning,
            )
        if not np.all((expr_upper >= exprv) | np.isnan(exprv)):
            warn(
                "data upper bounds are below the data values",
                category=QuantityOfInterestRuntimeWarning,
            )

        Xs_lower: np_sndarray[Ps, Ns, np.dtype[F]] = np.full(
            Xs.shape, Xs.dtype.type(-np.inf)
        )
        Xs_lower[np.isnan(Xs)] = np.nan
        Xs_lower[(...,) + self._index] = expr_lower

        Xs_upper: np_sndarray[Ps, Ns, np.dtype[F]] = np.full(
            Xs.shape, Xs.dtype.type(np.inf)
        )
        Xs_upper[np.isnan(Xs)] = np.nan
        Xs_upper[(...,) + self._index] = expr_upper

        return Xs_lower, Xs_upper

    @override
    def __repr__(self) -> str:
        if self._index == ():
            return "x"
        return f"X[{','.join(str(i) for i in self._index)}]"


Data.SCALAR = Data(index=())


class LateBoundConstant(EmptyExpr):
    __slots__: tuple[str, ...] = ("_name", "_index")
    _name: Parameter
    _index: tuple[int, ...]

    def __init__(self, name: Parameter, index: tuple[int, ...]) -> None:
        self._name = name
        self._index = index

    @staticmethod
    def like(name: Parameter, data: Data) -> "LateBoundConstant":
        return LateBoundConstant(name, data.index)

    @property
    def name(self) -> Parameter:
        return self._name

    @property
    def index(self) -> tuple[int, ...]:
        return self._index

    @property
    @override
    def args(self) -> tuple[()]:
        return ()

    @override
    def with_args(self) -> "LateBoundConstant":
        return LateBoundConstant(self._name, self._index)

    @override  # type: ignore
    def apply_array_element_offset(
        self,
        axis: int,
        offset: int,
    ) -> "LateBoundConstant":
        index = list(self._index)
        index[axis] += offset
        return LateBoundConstant(self._name, index=tuple(index))

    @property  # type: ignore
    @override
    def late_bound_constants(self) -> frozenset[Parameter]:
        return frozenset([self.name])

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return self

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        out: np.ndarray[tuple[int, ...], np.dtype[F]] = late_bound[self.name][
            (...,) + self._index
        ]
        assert _is_of_shape(out, Xs.shape[:1])
        return out

    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        assert False, "late-bound constants have no data bounds"

    @override
    def __repr__(self) -> str:
        if self._index == ():
            return f'c["{self._name}"]'
        return f'C["{self._name}"][{",".join(str(i) for i in self._index)}]'


# scalar constant that tricks the QoIs by pretending to be data,
# which is useful for tests and fuzzing to produce specific data-like values
class ScalarAnyDataConstant(EmptyExpr):
    __slots__: tuple[str, ...] = ("_const",)
    _const: np.number

    def __init__(self, const: np.number) -> None:
        self._const = const

    @property
    @override
    def args(self) -> tuple[()]:
        return ()

    @override
    def with_args(self) -> "ScalarAnyDataConstant":
        return ScalarAnyDataConstant(self._const)

    @property  # type: ignore
    @override
    def has_data(self) -> bool:
        return True

    @override  # type: ignore
    def eval_has_data(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[np.bool]]:
        return _ones(Xs.shape[:1], dtype=np.dtype(np.bool))

    @property  # type: ignore
    @override
    def data_indices(self) -> frozenset[tuple[int, ...]]:
        return frozenset()

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return self

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        assert isinstance(self._const, Xs.dtype.type)
        return _broadcast_to(self._const, Xs.shape[:1])

    @override
    @data_bounds(DataBounds.infallible)
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        assert isinstance(self._const, Xs.dtype.type)

        if not np.all((expr_lower <= self._const) | np.isnan(self._const)):
            warn(
                "data lower bounds are above the any data constant",
                category=QuantityOfInterestRuntimeWarning,
            )
        if not np.all((expr_upper >= self._const) | np.isnan(self._const)):
            warn(
                "data upper bounds are below the any data constant",
                category=QuantityOfInterestRuntimeWarning,
            )

        Xs_lower: np_sndarray[Ps, Ns, np.dtype[F]] = np.full(
            Xs.shape, Xs.dtype.type(-np.inf)
        )
        Xs_lower[np.isnan(Xs)] = np.nan

        Xs_upper: np_sndarray[Ps, Ns, np.dtype[F]] = np.full(
            Xs.shape, Xs.dtype.type(np.inf)
        )
        Xs_upper[np.isnan(Xs)] = np.nan

        return Xs_lower, Xs_upper

    @override
    def __repr__(self) -> str:
        return f"any_data({self._const!r})"



from collections.abc import Mapping

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import (
    _ensure_array,
    _is_sign_negative_number,
    _maximum_zero_sign_sensitive,
    _minimum_zero_sign_sensitive,
)
from ....utils.bindings import Parameter
from ..bound import checked_data_bounds, guarantee_arg_within_expr_bounds
from ..typing import F, Ns, Ps, np_sndarray
from .abc import AnyExpr, Expr
from .constfold import ScalarFoldedConstant


class ScalarSqrt(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr):
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarSqrt":
        return ScalarSqrt(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a, dtype, np.sqrt, ScalarSqrt
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.sqrt(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # for sqrt(-0.0), we should return -0.0 as the inverse
        # this ensures that 1/sqrt(-0.0) doesn't become 1/sqrt(0.0)
        def _sqrt_inv(
            x: np.ndarray[tuple[Ps], np.dtype[F]],
        ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
            out: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
                np.square(_maximum_zero_sign_sensitive(Xs.dtype.type(0), x))
            )
            np.copyto(out, x, where=(x == 0), casting="no")
            return out

        # evaluate arg and sqrt(arg)
        arg = self._a
        argv = arg.eval(Xs, late_bound)
        exprv = np.sqrt(argv)

        smallest_subnormal = np.finfo(Xs.dtype).smallest_subnormal

        # apply the inverse function to get the bounds on arg
        # sqrt(-0.0) = -0.0 and sqrt(+0.0) = +0.0
        # sqrt(...) is NaN for negative values and can then take any negative
        #  value
        # otherwise ensure that the bounds on sqrt(...) are non-negative
        # if arg_lower == argv and argv == -0.0, we need to guarantee that
        #  arg_lower is also -0.0, same for arg_upper
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            _minimum_zero_sign_sensitive(argv, _sqrt_inv(expr_lower))
        )
        arg_lower[np.less(argv, 0)] = -np.inf

        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            _maximum_zero_sign_sensitive(argv, _sqrt_inv(expr_upper))
        )
        arg_upper[np.less(argv, 0)] = -smallest_subnormal

        # we need to force argv if expr_lower == expr_upper
        np.copyto(arg_lower, argv, where=(expr_lower == expr_upper), casting="no")
        np.copyto(arg_upper, argv, where=(expr_lower == expr_upper), casting="no")

        # handle rounding errors in sqrt(square(...)) early
        arg_lower = guarantee_arg_within_expr_bounds(
            lambda arg_lower: np.sqrt(arg_lower),
            exprv,
            argv,
            arg_lower,
            expr_lower,
            expr_upper,
        )
        arg_upper = guarantee_arg_within_expr_bounds(
            lambda arg_upper: np.sqrt(arg_upper),
            exprv,
            argv,
            arg_upper,
            expr_lower,
            expr_upper,
        )

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"sqrt({self._a!r})"


class ScalarSquare(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr):
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarSquare":
        return ScalarSquare(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a, dtype, np.square, ScalarSquare
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.square(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg and square(arg)
        arg = self._a
        argv = arg.eval(Xs, late_bound)
        exprv = np.square(argv)

        # apply the inverse function to get the bounds on arg
        al = np.sqrt(_maximum_zero_sign_sensitive(expr_lower, Xs.dtype.type(0)))
        au = np.sqrt(expr_upper)

        # flip and swap the expr bounds to get the bounds on arg
        # square(...) cannot be negative, but
        #  - a > 0 and 0 < el <= eu -> al = el, au = eu
        #  - a < 0 and 0 < el <= eu -> al = -eu, au = -el
        #  - el <= 0 -> al = -eu, au = eu
        # TODO: an interval union could represent that the two sometimes-
        #       disjoint intervals in the future
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(al, copy=True)
        np.negative(
            au,
            out=arg_lower,
            where=(np.less_equal(expr_lower, 0) | _is_sign_negative_number(argv)),
        )
        arg_lower = _ensure_array(_minimum_zero_sign_sensitive(argv, arg_lower))

        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(au, copy=True)
        np.negative(
            al,
            out=arg_upper,
            where=(np.greater(expr_lower, 0) & _is_sign_negative_number(argv)),
        )
        arg_upper = _ensure_array(_maximum_zero_sign_sensitive(argv, arg_upper))

        # we need to force argv if expr_lower == expr_upper
        np.copyto(arg_lower, argv, where=(expr_lower == expr_upper), casting="no")
        np.copyto(arg_upper, argv, where=(expr_lower == expr_upper), casting="no")

        # handle rounding errors in square(sqrt(...)) early
        arg_lower = guarantee_arg_within_expr_bounds(
            lambda arg_lower: np.square(arg_lower),
            exprv,
            argv,
            arg_lower,
            expr_lower,
            expr_upper,
        )
        arg_upper = guarantee_arg_within_expr_bounds(
            lambda arg_upper: np.square(arg_upper),
            exprv,
            argv,
            arg_upper,
            expr_lower,
            expr_upper,
        )

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"square({self._a!r})"



from abc import ABC, abstractmethod
from collections.abc import Mapping

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils.bindings import Parameter
from ..bound import DataBounds, data_bounds
from ..typing import F, Ns, Ps, np_sndarray
from .abc import AnyExpr, Expr
from .literal import Number


class Reporter(ABC):
    """
    Abstract base class for reporters that are informed during quantity of
    interest expression computations.
    """

    __slots__: tuple[str, ...] = ()

    @abstractmethod
    def enter(self, expr: AnyExpr) -> None:
        """
        Report that `expr` has been entered.

        Parameters
        ----------
        expr : Expr
            The expression that has been entered.
        """

    @abstractmethod
    def exit(self, expr: AnyExpr) -> None:
        """
        Report that `expr` has been exited.

        Parameters
        ----------
        expr : Expr
            The expression that has been exited.
        """


class ReportingExpr(Expr[AnyExpr]):
    """
    A reporting expression wraps around an existing `expr`, forwarding all
    functionality to it while also informing the `reporter`.

    Parameters
    ----------
    expr : AnyExpr
        The expression that will be wrapped and reported on.
    reporter : Reporter
        The reporter that will report on the expression.
    """

    __slots__: tuple[str, ...] = ("_expr", "_reporter")
    _expr: AnyExpr
    _reporter: Reporter

    def __init__(self, expr: AnyExpr, reporter: Reporter) -> None:
        pass

    def __new__(cls, expr: AnyExpr, reporter: Reporter) -> "ReportingExpr | Number":  # type: ignore[misc]
        if isinstance(expr, ReportingExpr | Number):
            return expr
        # we have to also assign inside __new__ since the double ReportingExpr
        #  unwrapping would still call __init__ and create a reference loop
        this = super().__new__(cls)
        this._expr = expr
        this._reporter = reporter
        return this

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._expr,)

    @override
    def with_args(self, expr: AnyExpr) -> "ReportingExpr | Number":
        return ReportingExpr(expr, self._reporter)

    @property  # type: ignore[misc]
    @override
    def expr_size(self) -> int:
        return self._expr.expr_size

    @property  # type: ignore[misc]
    @override
    def data_expr_size(self) -> int:
        return self._expr.data_expr_size

    @property  # type: ignore[misc]
    @override
    def has_data(self) -> bool:
        return self._expr.has_data

    @override  # type: ignore
    def eval_has_data(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[np.bool]]:
        return self._expr.eval_has_data(Xs, late_bound)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        fexpr = self._expr.constant_fold(dtype)
        # partially / not constant folded -> stop further folding
        if isinstance(fexpr, Expr):
            return ReportingExpr(fexpr, self._reporter)
        # fully constant folded -> allow further folding
        return fexpr

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return self._expr.eval(Xs, late_bound)

    @data_bounds(DataBounds.infallible)
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        self._reporter.enter(self._expr)
        try:
            return self._expr.compute_data_bounds(
                expr_lower, expr_upper, Xs, late_bound
            )
        finally:
            self._reporter.exit(self._expr)

    @override
    def __repr__(self) -> str:
        return repr(self._expr)



from collections.abc import Mapping

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import _ensure_array, _is_sign_negative_number
from ....utils.bindings import Parameter
from ..bound import checked_data_bounds
from ..typing import F, Ns, Ps, np_sndarray
from .abc import AnyExpr, Expr
from .constfold import ScalarFoldedConstant


class ScalarAbs(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr) -> None:
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarAbs":
        return ScalarAbs(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a, dtype, np.abs, ScalarAbs
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.abs(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg
        arg = self._a
        argv = arg.eval(Xs, late_bound)

        # flip and swap the expr bounds to get the bounds on arg
        # abs(...) cannot be negative, but
        #  - a > 0 and 0 < el <= eu -> al = el, au = eu
        #  - a < 0 and 0 < el <= eu -> al = -eu, au = -el
        #  - el <= 0 -> al = -eu, au = eu
        # TODO: an interval union could represent that the two sometimes-
        #       disjoint intervals in the future
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            expr_lower, copy=True
        )
        np.negative(
            expr_upper,
            out=arg_lower,
            where=(np.less_equal(expr_lower, 0) | _is_sign_negative_number(argv)),
        )
        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            expr_upper, copy=True
        )
        np.negative(
            expr_lower,
            out=arg_upper,
            where=(np.greater(expr_lower, 0) & _is_sign_negative_number(argv)),
        )

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"abs({self._a!r})"



"""
Pointwise quantity of interest (QoI) error bound safeguard.
"""

__all__ = ["PointwiseQuantityOfInterestErrorBoundSafeguard"]

from collections.abc import Set
from typing import ClassVar, Literal

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import _ensure_array, _logical_and, _ones, _place, _reshape
from ....utils.bindings import Bindings, Parameter
from ....utils.cast import ToFloatMode, saturating_finite_float_cast, to_float
from ....utils.error import TypeCheckError, ctx, lookup_enum_or_raise
from ....utils.intervals import IntervalUnion
from ....utils.typing import JSON, F, S, T
from ..._qois import PointwiseQuantityOfInterest
from ..._qois.interval import compute_safe_data_lower_upper_interval_union
from ...eb import (
    ErrorBound,
    _apply_finite_qoi_error_bound,
    _check_error_bound,
    _compute_finite_absolute_error,
    _compute_finite_absolute_error_bound,
)
from ...qois import PointwiseQuantityOfInterestExpression
from ..abc import PointwiseSafeguard


class PointwiseQuantityOfInterestErrorBoundSafeguard(PointwiseSafeguard):
    """
    The `PointwiseQuantityOfInterestErrorBoundSafeguard` guarantees that the
    pointwise error `type` on a derived pointwise quantity of interest (QoI)
    is less than or equal to the provided bound `eb`.

    The quantity of interest is specified as a non-constant expression, in
    string form, over the pointwise value `x`. For example, to bound the error
    on the square of `x`, set `qoi="square(x)"` (or `qoi="x**2"`).

    If the derived quantity of interest for an element evaluates to an infinite
    value, this safeguard guarantees that the quantity of interest on the
    corrected value produces the exact same infinite value. For a NaN quantity
    of interest, this safeguard guarantees that the quantity of interest on the
    corrected value is also NaN, but does not guarantee that it has the same
    bit pattern.

    The error bound can be verified by evaluating the QoI in the floating-point
    data type selected by `qoi_dtype` parameter using the
    [`evaluate_qoi`][.evaluate_qoi] method.

    Please refer to the
    [`PointwiseQuantityOfInterestExpression`][.....qois.PointwiseQuantityOfInterestExpression]
    for the EBNF grammar that specifies the language in which the quantities of
    interest are written.

    The implementation was originally inspired by:

    > Pu Jiao, Sheng Di, Hanqi Guo, Kai Zhao, Jiannan Tian, Dingwen Tao, Xin
    Liang, and Franck Cappello. (2022). Toward Quantity-of-Interest Preserving
    Lossy Compression for Scientific Data. *Proceedings of the VLDB Endowment*.
    16, 4 (December 2022), 697-710. Available from:
    [doi:10.14778/3574245.3574255](https://doi.org/10.14778/3574245.3574255).

    Parameters
    ----------
    qoi : PointwiseExpr
        The non-constant expression for computing the derived quantity of
        interest over a pointwise value `x`.
    type : str | ErrorBound
        The type of error bound on the quantity of interest that is enforced by
        this safeguard.
    eb : int | float | str | Parameter
        The value of or late-bound parameter name for the error bound on the
        quantity of interest that is enforced by this safeguard.

        The error bound is applied relative to the values of the quantity of
        interest evaluated on the original data.
    qoi_dtype : str | ToFloatMode
        The floating-point data type in which the quantity of interest is
        evaluated. By default, the smallest floating-point data type that can
        losslessly represent all input data values is chosen.

    Raises
    ------
    TypeCheckError
        if any parameter has the wrong type.
    SyntaxError
        if the `qoi` is not a valid pointwise quantity of interest expression.
    ValueError
        if `type` does not name a valid error bound, or the `qoi_dtype` does
        not name a valid floating-point data type.
    ValueError
        if `eb` is an invalid error bound value for the error bound `type`.
    """

    __slots__: tuple[str, ...] = (
        "_qoi",
        "_type",
        "_eb",
        "_qoi_dtype",
        "_qoi_expr",
    )
    _qoi: PointwiseQuantityOfInterestExpression
    _type: ErrorBound
    _eb: int | float | Parameter
    _qoi_dtype: ToFloatMode
    _qoi_expr: PointwiseQuantityOfInterest

    kind: ClassVar[str] = "qoi_eb_pw"

    def __init__(
        self,
        qoi: PointwiseQuantityOfInterestExpression,
        type: str | ErrorBound,
        eb: int | float | str | Parameter,
        qoi_dtype: str | ToFloatMode = ToFloatMode.lossless,
    ) -> None:
        with ctx.safeguard(self):
            with ctx.parameter("qoi"):
                TypeCheckError.check_instance_or_raise(qoi, str)

            with ctx.parameter("type"):
                TypeCheckError.check_instance_or_raise(type, str | ErrorBound)
                self._type = (
                    type
                    if isinstance(type, ErrorBound)
                    else lookup_enum_or_raise(ErrorBound, type)
                )

            with ctx.parameter("eb"):
                TypeCheckError.check_instance_or_raise(
                    eb, int | float | str | Parameter
                )
                if isinstance(eb, Parameter):
                    self._eb = eb
                elif isinstance(eb, str):
                    self._eb = Parameter(eb)
                else:
                    _check_error_bound(self._type, eb)
                    self._eb = eb

            with ctx.parameter("qoi_dtype"):
                TypeCheckError.check_instance_or_raise(qoi_dtype, str | ToFloatMode)
                self._qoi_dtype = (
                    qoi_dtype
                    if isinstance(qoi_dtype, ToFloatMode)
                    else lookup_enum_or_raise(ToFloatMode, qoi_dtype)
                )

            with ctx.parameter("qoi"):
                self._qoi = qoi
                self._qoi_expr = PointwiseQuantityOfInterest(qoi)

    @property
    @override
    def late_bound(self) -> Set[Parameter]:
        """
        The set of late-bound parameters that this safeguard has.

        Late-bound parameters are only bound when checking and applying the
        safeguard, in contrast to the normal early-bound parameters that are
        configured during safeguard initialisation.

        Late-bound parameters can be used for parameters that depend on the
        specific data that is to be safeguarded.
        """

        parameters = frozenset(self._qoi_expr.late_bound_constants)

        if isinstance(self._eb, Parameter):
            parameters = parameters.union([self._eb])

        return parameters

    def evaluate_qoi(
        self,
        data: np.ndarray[S, np.dtype[T]],
        late_bound: Bindings,
    ) -> np.ndarray[S, np.dtype[F]]:
        """
        Evaluate the derived quantity of interest on the `data` in the
        floating-point data type selected by the `qoi_dtype` parameter.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Data for which the quantity of interest is evaluated.
        late_bound : Bindings
            Bindings for late-bound constants in the quantity of interest.

        Returns
        -------
        qoi : np.ndarray[S, np.dtype[F]]
            Evaluated quantity of interest, in floating-point.

        Raises
        ------
        TypeError
            if the `data` could not be losslessly cast to `qoi_dtype`.
        LateBoundParameterResolutionError
            if any of the `qoi`'s late-bound constants is not contained in the
            bindings.
        ValueError
            if any late-bound constant could not be broadcast to the `data`'s
            shape.
        TypeError
            if any late-bound constant is floating-point but the `data` is
            integer.
        ValueError
            if not all values for all late-bound constants could be losslessly
            converted to the `data`'s type.
        """

        with ctx.safeguard(self):
            with ctx.parameter("qoi_dtype"):
                ftype: np.dtype[F] = self._qoi_dtype.floating_point_dtype_for(
                    data.dtype
                )  # type: ignore
                data_float: np.ndarray[tuple[int], np.dtype[F]] = to_float(
                    data, ftype=ftype
                ).flatten()

            late_bound_constants: dict[
                Parameter, np.ndarray[tuple[int], np.dtype[F]]
            ] = dict()
            with ctx.parameter("qoi"):
                for c in self._qoi_expr.late_bound_constants:
                    with ctx.parameter(c):
                        late_bound_constants[c] = (
                            late_bound.resolve_ndarray_with_lossless_cast(
                                c, data.shape, ftype
                            ).flatten()
                        )

        qoi_: np.ndarray[tuple[int], np.dtype[F]] = self._qoi_expr.eval(
            data_float,
            late_bound_constants,
        )
        qoi: np.ndarray[S, np.dtype[F]] = _reshape(qoi_, data.shape)
        return qoi

    @override
    def check_pointwise(
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Check which elements in the `prediction` array satisfy the error bound
        for the quantity of interest on the `data`.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Original data array, relative to which the `prediction` is checked.
        prediction : np.ndarray[S, np.dtype[T]]
            Prediction for the `data` array.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only check at data points where the condition is [`True`][True].

        Returns
        -------
        ok : np.ndarray[S, np.dtype[np.bool]]
            Pointwise, `True` if the check succeeded for this element.

        Raises
        ------
        TypeError
            if the `data` could not be losslessly cast to `qoi_dtype`.
        LateBoundParameterResolutionError
            if any of the `qoi`'s late-bound constants is not contained in the
            bindings.
        ValueError
            if any late-bound constant could not be broadcast to the `data`'s
            shape.
        TypeError
            if any late-bound constant is floating-point but the `data` is
            integer.
        ValueError
            if not all values for all late-bound constants could be losslessly
            converted to the `data`'s type.
        LateBoundParameterResolutionError
            if the error bound `eb` is late-bound but its late-bound parameter
            is not in `late_bound`.
        ValueError
            if the late-bound `eb` could not be broadcast to the `data`'s shape.
        ValueError
            if the late-bound `eb` is non-finite, i.e. infinite or NaN, or an
            invalid error bound value for the error bound `type`.
        """

        with ctx.safeguard(self):
            with ctx.parameter("qoi_dtype"):
                ftype: np.dtype[np.floating] = self._qoi_dtype.floating_point_dtype_for(
                    data.dtype
                )
                data_float: np.ndarray[tuple[int], np.dtype[np.floating]] = to_float(
                    data, ftype=ftype
                ).flatten()
                prediction_float: np.ndarray[tuple[int], np.dtype[np.floating]] = (
                    to_float(prediction, ftype=ftype).flatten()
                )

            late_bound_constants: dict[
                Parameter, np.ndarray[tuple[int], np.dtype[np.floating]]
            ] = dict()
            with ctx.parameter("qoi"):
                for c in self._qoi_expr.late_bound_constants:
                    with ctx.parameter(c):
                        late_bound_constants[c] = (
                            late_bound.resolve_ndarray_with_lossless_cast(
                                c, data.shape, ftype
                            ).flatten()
                        )

            # optimization: only evaluate the QoI where necessary
            if where is not True:
                data_float = np.extract(where, data_float)
                prediction_float = np.extract(where, prediction_float)
                late_bound_constants = {
                    c: np.extract(where, cv) for (c, cv) in late_bound_constants.items()
                }

            qoi_data: np.ndarray[tuple[int], np.dtype[np.floating]] = (
                self._qoi_expr.eval(data_float, late_bound_constants)
            )
            qoi_prediction: np.ndarray[tuple[int], np.dtype[np.floating]] = (
                self._qoi_expr.eval(prediction_float, late_bound_constants)
            )

            with ctx.parameter("eb"):
                eb: np.ndarray[tuple[()] | tuple[int], np.dtype[np.floating]] = (
                    late_bound.resolve_ndarray_with_saturating_finite_float_cast(
                        self._eb,
                        data.shape,
                        ftype,
                    ).flatten()
                    if isinstance(self._eb, Parameter)
                    else saturating_finite_float_cast(self._eb, ftype)
                )
                if isinstance(self._eb, Parameter):
                    _eb: Parameter = self._eb
                    with ctx.late_bound_parameter(_eb):
                        _check_error_bound(self._type, eb)

        if where is not True and eb.shape != ():
            eb = np.extract(where, eb)

        finite_ok: np.ndarray[tuple[int], np.dtype[np.bool]] = np.less_equal(
            _compute_finite_absolute_error(self._type, qoi_data, qoi_prediction),
            _compute_finite_absolute_error_bound(self._type, eb, qoi_data),
        )

        ok_: np.ndarray[tuple[int], np.dtype[np.bool]] = _ensure_array(finite_ok)
        np.equal(qoi_data, qoi_prediction, out=ok_, where=np.isinf(qoi_data))
        np.isnan(qoi_prediction, out=ok_, where=np.isnan(qoi_data))

        # the check succeeds where `where` is False
        ok: np.ndarray[S, np.dtype[np.bool]]
        if where is True:
            ok = _reshape(ok_, data.shape)
        else:
            ok = _ones(data.shape, np.dtype(np.bool))
            _place(ok, where, ok_)

        return ok

    @override
    def compute_safe_intervals(
        self,
        data: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> IntervalUnion[T, int, int]:
        """
        Compute the intervals in which the error bound is upheld with
        respect to the quantity of interest on the `data`.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Data for which the safe intervals should be computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the safe intervals at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        intervals : IntervalUnion[T, int, int]
            Union of intervals in which the error bound is upheld.

        Raises
        ------
        TypeError
            if the `data` could not be losslessly cast to `qoi_dtype`.
        LateBoundParameterResolutionError
            if any of the `qoi`'s late-bound constants is not contained in the
            bindings.
        ValueError
            if any late-bound constant could not be broadcast to the `data`'s
            shape.
        TypeError
            if any late-bound constant is floating-point but the `data` is
            integer.
        ValueError
            if not all values for all late-bound constants could be losslessly
            converted to the `data`'s type.
        LateBoundParameterResolutionError
            if the error bound `eb` is late-bound but its late-bound parameter
            is not in `late_bound`.
        ValueError
            if the late-bound `eb` could not be broadcast to the `data`'s shape.
        ValueError
            if the late-bound `eb` is non-finite, i.e. infinite or NaN, or an
            invalid error bound value for the error bound `type`.
        """

        with ctx.safeguard(self):
            with ctx.parameter("qoi_dtype"):
                ftype: np.dtype[np.floating] = self._qoi_dtype.floating_point_dtype_for(
                    data.dtype
                )
                data_float: np.ndarray[tuple[int], np.dtype[np.floating]] = to_float(
                    data, ftype=ftype
                ).flatten()

            late_bound_constants: dict[
                Parameter, np.ndarray[tuple[int], np.dtype[np.floating]]
            ] = dict()
            with ctx.parameter("qoi"):
                for c in self._qoi_expr.late_bound_constants:
                    with ctx.parameter(c):
                        late_bound_constants[c] = (
                            late_bound.resolve_ndarray_with_lossless_cast(
                                c, data.shape, ftype
                            ).flatten()
                        )

            # optimization: only evaluate the QoI and compute data bounds where
            #  necessary
            if where is not True:
                data_float = np.extract(where, data_float)
                late_bound_constants = {
                    c: np.extract(where, cv) for (c, cv) in late_bound_constants.items()
                }

            data_qoi: np.ndarray[tuple[int], np.dtype[np.floating]] = (
                self._qoi_expr.eval(data_float, late_bound_constants)
            )

            with ctx.parameter("eb"):
                eb: np.ndarray[tuple[()] | tuple[int], np.dtype[np.floating]] = (
                    late_bound.resolve_ndarray_with_saturating_finite_float_cast(
                        self._eb,
                        data.shape,
                        ftype,
                    ).flatten()
                    if isinstance(self._eb, Parameter)
                    else saturating_finite_float_cast(self._eb, ftype)
                )
                if isinstance(self._eb, Parameter):
                    _eb: Parameter = self._eb
                    with ctx.late_bound_parameter(_eb):
                        _check_error_bound(self._type, eb)

            if where is not True and eb.shape != ():
                eb = np.extract(where, eb)

        qoi_lower, qoi_upper = _apply_finite_qoi_error_bound(self._type, eb, data_qoi)

        # compute the bounds in data space
        data_float_lower_, data_float_upper_ = self._qoi_expr.compute_data_bounds(
            qoi_lower,
            qoi_upper,
            data_float,
            late_bound_constants,
        )

        # the data bounds can be arbitrary where `where` is False since they
        #  are later overridden by preserve_only_where
        data_float_lower: np.ndarray[S, np.dtype[np.floating]]
        data_float_upper: np.ndarray[S, np.dtype[np.floating]]
        if where is True:
            data_float_lower = _reshape(data_float_lower_, data.shape)
            data_float_upper = _reshape(data_float_upper_, data.shape)
        else:
            data_float_lower = np.full(data.shape, -np.inf, ftype)
            data_float_upper = np.full(data.shape, np.inf, ftype)
            _place(data_float_lower, where, data_float_lower_)
            _place(data_float_upper, where, data_float_upper_)

        wheref: Literal[True] | np.ndarray[tuple[int], np.dtype[np.bool]] = (
            True if where is True else where.flatten()
        )

        return compute_safe_data_lower_upper_interval_union(
            data, data_float_lower, data_float_upper
        ).preserve_only_where(wheref)

    @override
    def compute_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Compute the footprint of the `foot` array, e.g. for expanding data
        points into the pointwise checks that they contribute to.

        The footprint is equivalent to `foot & where`.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the footprint at pointwise checks where the condition
            is [`True`][True].

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The footprint of the `foot` array.
        """

        return _logical_and(foot, where)

    @override
    def compute_inverse_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Compute the inverse footprint of the `foot` array, e.g. for expanding
        pointwise check fails into the points that could have contributed to
        the failures.

        The inverse footprint is equivalent to `foot & where`.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the inverse footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the inverse footprint at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The inverse footprint of the `foot` array.
        """

        return _logical_and(foot, where)

    @override
    def get_config(self) -> dict[str, JSON]:
        """
        Returns the configuration of the safeguard.

        Returns
        -------
        config : dict[str, JSON]
            Configuration of the safeguard.
        """

        return dict(
            kind=type(self).kind,
            qoi=self._qoi,
            type=self._type.name,
            eb=self._eb,
            qoi_dtype=self._qoi_dtype.name,
        )



__all__ = ["PointwiseQuantityOfInterest", "StencilQuantityOfInterest"]

from collections.abc import Mapping, Set

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ...utils._compat import _ensure_array, _zeros
from ...utils.bindings import Parameter
from ...utils.error import ctx
from ..qois import (
    PointwiseQuantityOfInterestExpression,
    StencilQuantityOfInterestExpression,
)
from .expr.abc import AnyExpr, Expr
from .expr.array import Array
from .expr.constfold import ScalarFoldedConstant
from .expr.data import Data
from .lexer import QoILexer
from .parser import QoIParser
from .typing import F, Ns, Ps, coerce_to_flat, np_sndarray


class PointwiseQuantityOfInterest:
    """
    Pointwise quantity of interest, which handles parsing, evaluation, and
    data bound propagation for the QoI expression.

    Parameters
    ----------
    qoi : PointwiseQuantityOfInterestExpression
        The pointwise quantity of interest in [`str`][str]ing form.

    Raises
    ------
    SyntaxError
        if the `qoi` is not a valid pointwise quantity of interest expression.
    """

    __slots__: tuple[str, ...] = ("_expr", "_late_bound_constants")
    _expr: AnyExpr
    _late_bound_constants: frozenset[Parameter]

    def __init__(self, qoi: PointwiseQuantityOfInterestExpression):
        lexer = QoILexer()
        parser = QoIParser(x=Data.SCALAR, X=None, I=None)

        expr = parser.parse2(qoi, lexer.tokenize(qoi))
        assert isinstance(expr, Expr)

        if isinstance(expr, Array):
            raise (
                SyntaxError(
                    "QoI expression must evaluate to a scalar, not an array "
                    + f"expression of shape {expr.shape}",
                    ("<qoi>", None, None, None),
                )
                | ctx
            )

        if not expr.has_data:
            raise (
                SyntaxError(
                    "QoI expression must not be constant", ("<qoi>", None, None, None)
                )
                | ctx
            )

        late_bound_constants = expr.late_bound_constants

        dummy_pointwise: np.ndarray[tuple[int], np.dtype[np.float64]] = _zeros(
            (0,), np.dtype(np.float64)
        )

        with np.errstate(
            divide="ignore", over="ignore", under="ignore", invalid="ignore"
        ):
            # check if the expression is well-formed and if data bounds can be
            #  computed
            _canary_expr = expr.constant_fold(np.dtype(np.float64))
            if isinstance(_canary_expr, Expr):
                _canary_data_bounds = _canary_expr.compute_data_bounds(
                    dummy_pointwise,
                    dummy_pointwise,
                    dummy_pointwise,
                    {c: dummy_pointwise for c in late_bound_constants},
                )

        self._expr = expr
        self._late_bound_constants = late_bound_constants

    @property
    def late_bound_constants(self) -> Set[Parameter]:
        """
        The set of late-bound constant parameters that this QoI uses.
        """

        return self._late_bound_constants

    @np.errstate(divide="ignore", over="ignore", under="ignore", invalid="ignore")
    def eval(
        self,
        X: np.ndarray[tuple[Ps], np.dtype[F]],
        late_bound: Mapping[Parameter, np.ndarray[tuple[Ps], np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        """
        Evaluate this pointwise quantity of interest on the data `X`.

        Parameters
        ----------
        X : np.ndarray[tuple[Ps], np.dtype[F]]
            The pointwise data, in floating-point format.
        late_bound : Mapping[Parameter, np.ndarray[tuple[Ps], np.dtype[F]]]
            The late-bound constants parameters for this QoI, with the same
            shape and floating-point dtype as the data.

        Returns
        -------
        qoi : np.ndarray[tuple[Ps], np.dtype[F]]
            The pointwise quantity of interest values.
        """

        X = _ensure_array(X)
        expr = ScalarFoldedConstant.constant_fold_expr(self._expr, X.dtype)
        exprv = _ensure_array(expr.eval(X, late_bound))
        assert exprv.dtype == X.dtype
        assert exprv.shape == X.shape
        return exprv

    @np.errstate(divide="ignore", over="ignore", under="ignore", invalid="ignore")
    def compute_data_bounds(
        self,
        qoi_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        qoi_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        X: np.ndarray[tuple[Ps], np.dtype[F]],
        late_bound: Mapping[Parameter, np.ndarray[tuple[Ps], np.dtype[F]]],
    ) -> tuple[np.ndarray[tuple[Ps], np.dtype[F]], np.ndarray[tuple[Ps], np.dtype[F]]]:
        """
        Compute the lower-upper bounds on the data `X` that satisfy the
        lower-upper bounds `qoi_lower` and `qoi_lower` on the QoI.

        Parameters
        ----------
        qoi_lower : np.ndarray[tuple[Ps], np.dtype[F]]
            The pointwise lower bound on the QoI.
        qoi_upper : np.ndarray[tuple[Ps], np.dtype[F]]
            The pointwise upper bound on the QoI.
        X : np.ndarray[tuple[Ps], np.dtype[F]]
            The pointwise data, in floating-point format.
        late_bound : Mapping[Parameter, np.ndarray[tuple[Ps], np.dtype[F]]]
            The late-bound constants parameters for this QoI, with the same
            shape and floating-point dtype as the data.

        Returns
        -------
        X_lower, X_upper : tuple[np.ndarray[tuple[Ps], np.dtype[F]], np.ndarray[tuple[Ps], np.dtype[F]]]
            The pointwise lower and upper bounds on the data `X`.
        """

        qoi_lower = _ensure_array(qoi_lower)
        qoi_upper = _ensure_array(qoi_upper)
        X = _ensure_array(X)
        X_lower: np.ndarray[tuple[Ps], np.dtype[F]]
        X_upper: np.ndarray[tuple[Ps], np.dtype[F]]
        expr = self._expr.constant_fold(X.dtype)
        if isinstance(expr, Expr):
            X_lower_upper: tuple[
                np_sndarray[Ps, tuple[()], np.dtype[F]],
                np_sndarray[Ps, tuple[()], np.dtype[F]],
            ] = expr.compute_data_bounds(qoi_lower, qoi_upper, X, late_bound)
            X_lower = _ensure_array(coerce_to_flat(X_lower_upper[0]))
            X_upper = _ensure_array(coerce_to_flat(X_lower_upper[1]))
        else:
            # constant fold with combinators can create top-level folded consts
            X_lower = np.full(X.shape, X.dtype.type(-np.inf))
            X_lower[np.isnan(X)] = np.nan
            X_upper = np.full(X.shape, X.dtype.type(np.inf))
            X_upper[np.isnan(X)] = np.nan
        assert X_lower.dtype == X.dtype
        assert X_upper.dtype == X.dtype
        assert X_lower.shape == X.shape
        assert X_upper.shape == X.shape
        return X_lower, X_upper

    @override
    def __repr__(self) -> str:
        return repr(self._expr)


class StencilQuantityOfInterest:
    """
    Stencil quantity of interest, which handles parsing, evaluation, and
    data bound propagation for the QoI expression.

    Parameters
    ----------
    qoi : StencilQuantityOfInterestExpression
        The stencil quantity of interest in [`str`][str]ing form.
    stencil_shape : tuple[int, ...]
        The shape of the stencil neighbourhood.
    stencil_I : tuple[int, ...]
        The index `I` for the centre of the stencil neighbourhood.

    Raises
    ------
    SyntaxError
        if the `qoi` is not a valid stencil quantity of interest expression.
    """

    __slots__: tuple[str, ...] = (
        "_expr",
        "_stencil_shape",
        "_stencil_I",
        "_late_bound_constants",
    )
    _expr: AnyExpr
    _stencil_shape: tuple[int, ...]
    _stencil_I: tuple[int, ...]
    _late_bound_constants: frozenset[Parameter]

    def __init__(
        self,
        qoi: StencilQuantityOfInterestExpression,
        stencil_shape: tuple[int, ...],
        stencil_I: tuple[int, ...],
    ):
        assert len(stencil_shape) == len(stencil_I)
        assert all(s > 0 for s in stencil_shape)
        assert all(i >= 0 and i < s for i, s in zip(stencil_I, stencil_shape))

        self._stencil_shape = stencil_shape
        self._stencil_I = stencil_I

        lexer = QoILexer()
        parser = QoIParser(
            x=Data(index=stencil_I), X=Array.from_data_shape(stencil_shape), I=stencil_I
        )

        expr = parser.parse2(qoi, lexer.tokenize(qoi))
        assert isinstance(expr, Expr)

        if isinstance(expr, Array):
            raise (
                SyntaxError(
                    "QoI expression must evaluate to a scalar, not an array "
                    + f"expression of shape {expr.shape}",
                    ("<qoi>", None, None, None),
                )
                | ctx
            )

        if not expr.has_data:
            raise (
                SyntaxError(
                    "QoI expression must not be constant", ("<qoi>", None, None, None)
                )
                | ctx
            )

        late_bound_constants = expr.late_bound_constants

        dummy_pointwise: np.ndarray[tuple[int], np.dtype[np.float64]] = _zeros(
            (0,), np.dtype(np.float64)
        )
        dummy_stencil: np_sndarray[int, tuple[int, ...], np.dtype[np.float64]] = _zeros(
            (0, *stencil_shape), np.dtype(np.float64)
        )

        with np.errstate(
            divide="ignore", over="ignore", under="ignore", invalid="ignore"
        ):
            # check if the expression is well-formed and if data bounds can be
            #  computed
            _canary_expr = expr.constant_fold(np.dtype(np.float64))
            if isinstance(_canary_expr, Expr):
                _canary_data_bounds = _canary_expr.compute_data_bounds(
                    dummy_pointwise,
                    dummy_pointwise,
                    dummy_stencil,
                    {c: dummy_stencil for c in late_bound_constants},
                )

        self._expr = expr
        self._late_bound_constants = late_bound_constants

    @property
    def late_bound_constants(self) -> Set[Parameter]:
        """
        The set of late-bound constant parameters that this QoI uses.
        """

        return self._late_bound_constants

    @property
    def data_indices(self) -> Set[tuple[int, ...]]:
        """
        The set of data stencil indices `X[is]` that this QoI uses.
        """

        return self._expr.data_indices

    @np.errstate(divide="ignore", over="ignore", under="ignore", invalid="ignore")
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        """
        Evaluate this stencil quantity of interest on the stencil-extended data
        `Xs`.

        Parameters
        ----------
        Xs : np_sndarray[Ps, Ns, np.dtype[F]]
            The stencil-extended data, in floating-point format, which must be
            of shape [Ps, *stencil_shape].
        late_bound : Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]]
            The late-bound constants parameters for this QoI, with the same
            shape and floating-point dtype as the stencil-extended data.

        Returns
        -------
        qoi : np.ndarray[tuple[Ps], np.dtype[F]]
            The pointwise quantity of interest values.
        """

        Xs = _ensure_array(Xs)
        assert Xs.shape[1:] == self._stencil_shape
        expr = ScalarFoldedConstant.constant_fold_expr(self._expr, Xs.dtype)
        exprv = _ensure_array(expr.eval(Xs, late_bound))
        assert exprv.dtype == Xs.dtype
        assert exprv.shape == Xs.shape[:1]
        return exprv

    @np.errstate(divide="ignore", over="ignore", under="ignore", invalid="ignore")
    def compute_data_bounds(
        self,
        qoi_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        qoi_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        """
        Compute the lower-upper bounds on the stencil-extended data `Xs` that
        satisfy the lower-upper bounds `qoi_lower` and `qoi_lower` on the QoI.

        Parameters
        ----------
        qoi_lower : np.ndarray[tuple[Ps], np.dtype[F]]
            The pointwise lower bound on the QoI.
        qoi_upper : np.ndarray[tuple[Ps], np.dtype[F]]
            The pointwise upper bound on the QoI.
        Xs : np_sndarray[Ps, Ns, np.dtype[F]]
            The stencil-extended data, in floating-point format, which must be
            of shape [Ps, *stencil_shape].
        late_bound : Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]]
            The late-bound constants parameters for this QoI, with the same
            shape and floating-point dtype as the stencil-extended data.

        Returns
        -------
        Xs_lower, Xs_upper : tuple[np_sndarray[Ps, Ns, np.dtype[F]], np_sndarray[Ps, Ns, np.dtype[F]]]
            The stencil-extended lower and upper bounds on the stencil-extended
            data `Xs`.

            The bounds have not yet been combined across neighbouring points
            that contribute to the same QoI points.
        """

        qoi_lower = _ensure_array(qoi_lower)
        qoi_upper = _ensure_array(qoi_upper)
        Xs = _ensure_array(Xs)
        assert Xs.shape[:1] == qoi_lower.shape
        assert Xs.shape[1:] == self._stencil_shape
        Xs_lower: np_sndarray[Ps, Ns, np.dtype[F]]
        Xs_upper: np_sndarray[Ps, Ns, np.dtype[F]]
        expr = self._expr.constant_fold(Xs.dtype)
        if isinstance(expr, Expr):
            Xs_lower, Xs_upper = expr.compute_data_bounds(
                qoi_lower, qoi_upper, Xs, late_bound
            )
            Xs_lower = _ensure_array(Xs_lower)
            Xs_upper = _ensure_array(Xs_upper)
        else:
            # constant fold with combinators can create top-level folded consts
            Xs_lower = np.full(Xs.shape, Xs.dtype.type(-np.inf))
            Xs_lower[np.isnan(Xs)] = np.nan
            Xs_upper = np.full(Xs.shape, Xs.dtype.type(np.inf))
            Xs_upper[np.isnan(Xs)] = np.nan
        assert Xs_lower.dtype == Xs.dtype
        assert Xs_upper.dtype == Xs.dtype
        assert Xs_lower.shape == Xs.shape
        assert Xs_upper.shape == Xs.shape
        return Xs_lower, Xs_upper

    @override
    def __repr__(self) -> str:
        return repr(self._expr)



"""
float128, a floating-point dtype for numpy with true 128bit precision.
"""

__all__ = [
    "_float128",
    "_float128_type",
    "_float128_dtype",
    "_float128_pi",
    "_float128_e",
]

from collections.abc import Callable
from typing import TYPE_CHECKING, TypeAlias

import numpy as np
import numpy_quaddtype
from numpy._typing import _128Bit

if TYPE_CHECKING:
    _float128_type: TypeAlias = np.floating[_128Bit]
else:
    _float128_type: type[np.floating[_128Bit]] = numpy_quaddtype.QuadPrecision

_float128: Callable[[int | float | str], _float128_type] = (
    numpy_quaddtype.SleefQuadPrecision
)
_float128_dtype: np.dtype[_float128_type] = numpy_quaddtype.SleefQuadPrecDType()
_float128_pi: _float128_type = numpy_quaddtype.pi
_float128_e: _float128_type = numpy_quaddtype.e



"""
Logical all (and) combinator safeguard.
"""

__all__ = ["AllSafeguards"]

from abc import ABC
from collections.abc import Collection, Set
from typing import ClassVar, Literal

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ...utils.bindings import Bindings, Parameter
from ...utils.error import TypeCheckError, ctx
from ...utils.intervals import IntervalUnion
from ...utils.typing import JSON, S, T
from ..abc import Safeguard
from ..pointwise.abc import PointwiseSafeguard
from ..stencil import BoundaryCondition, NeighbourhoodAxis
from ..stencil.abc import StencilSafeguard


class AllSafeguards(Safeguard):
    """
    The `AllSafeguards` guarantees that, for each element, all of the combined
    safeguards' guarantees are upheld.

    At the moment, only pointwise and stencil safeguards and combinations
    thereof can be combined by this all-combinator. The combinator is a
    pointwise or a stencil safeguard, depending on the safeguards it combines.

    Parameters
    ----------
    safeguards : Collection[dict[str, JSON] | PointwiseSafeguard | StencilSafeguard]
        At least one safeguard configuration [`dict`][dict]s or already
        initialized [`PointwiseSafeguard`][....pointwise.abc.PointwiseSafeguard]
        or [`StencilSafeguard`][....stencil.abc.StencilSafeguard].

    Raises
    ------
    TypeCheckError
        if any parameter has the wrong type.
    ValueError
        if the `safeguards` collection is empty.
    ...
        if instantiating a safeguard raises an exception.
    """

    __slots__: tuple[str, ...] = ()

    kind: ClassVar[str] = "all"

    def __init__(
        self,
        *,
        safeguards: Collection[dict[str, JSON] | PointwiseSafeguard | StencilSafeguard],
    ) -> None:
        pass

    def __new__(  # type: ignore
        cls,
        *,
        safeguards: Collection[dict[str, JSON] | PointwiseSafeguard | StencilSafeguard],
    ) -> "_AllPointwiseSafeguards | _AllStencilSafeguards":
        from ... import SafeguardKind  # noqa: PLC0415

        with ctx.safeguardty(cls):
            with ctx.parameter("safeguards"):
                TypeCheckError.check_instance_or_raise(safeguards, Collection)

                if len(safeguards) <= 0:
                    raise (
                        ValueError("can only combine over at least one safeguard") | ctx
                    )

                safeguards_: list[PointwiseSafeguard | StencilSafeguard] = []
                safeguard: dict[str, JSON] | Safeguard
                for i, safeguard in enumerate(safeguards):
                    with ctx.index(i):
                        TypeCheckError.check_instance_or_raise(
                            safeguard, dict | PointwiseSafeguard | StencilSafeguard
                        )
                        if isinstance(safeguard, dict):
                            safeguard = SafeguardKind.from_config(safeguard)
                        TypeCheckError.check_instance_or_raise(
                            safeguard, PointwiseSafeguard | StencilSafeguard
                        )
                        safeguards_.append(safeguard)  # type: ignore

        pointwise_safeguards_: list[PointwiseSafeguard] = [
            safeguard
            for safeguard in safeguards_
            if isinstance(safeguard, PointwiseSafeguard)
        ]

        if len(pointwise_safeguards_) == len(safeguards_):
            return _AllPointwiseSafeguards(*pointwise_safeguards_)
        else:
            return _AllStencilSafeguards(*safeguards_)

    @property
    def safeguards(self) -> tuple[PointwiseSafeguard | StencilSafeguard, ...]:  # type: ignore
        """
        The set of safeguards that this any combinator has been configured to
        uphold.
        """

        ...

    @property
    @override
    def late_bound(self) -> Set[Parameter]:  # type: ignore
        """
        The set of late-bound parameters that this safeguard has.

        Late-bound parameters are only bound when checking and applying the
        safeguard, in contrast to the normal early-bound parameters that are
        configured during safeguard initialisation.

        Late-bound parameters can be used for parameters that depend on the
        specific data that is to be safeguarded.
        """

        ...

    @override
    def check(  # type: ignore
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> bool:
        """
        Check if, for all elements, all of the combined safeguards succeed the
        check.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Original data array, relative to which the `prediction` is checked.
        prediction : np.ndarray[S, np.dtype[T]]
            Prediction for the `data` array.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only check at data points where the condition is [`True`][True].

        Returns
        -------
        ok : bool
            `True` if the check succeeded.

        Raises
        ------
        ...
            if checking a safeguard raises an exception.
        """

        ...

    def check_pointwise(  # type: ignore
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Check for which elements all of the combined safeguards succeed the
        check.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Original data array, relative to which the `prediction` is checked.
        prediction : np.ndarray[S, np.dtype[T]]
            Prediction for the `data` array.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only check at data points where the condition is [`True`][True].

        Returns
        -------
        ok : np.ndarray[S, np.dtype[np.bool]]
            Pointwise, `True` if the check succeeded for this element.

        Raises
        ------
        ...
            if checking a safeguard raises an exception.
        """

        ...

    def compute_safe_intervals(  # type: ignore
        self,
        data: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> IntervalUnion[T, int, int]:
        """
        Compute the intersection of the safe intervals of the combined
        safeguards, i.e. where all of them are safe.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Data for which the safe intervals should be computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the safe intervals at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        intervals : IntervalUnion[T, int, int]
            Intersection of safe intervals.

        Raises
        ------
        ...
            if computing the safe intervals for a safeguard raises an exception.
        """

        ...

    def compute_footprint(  # type: ignore
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Compute the footprint of the `foot` array, e.g. for expanding data
        points into the pointwise checks that they contribute to.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the inverse footprint at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The footprint of the `foot` array.
        """

        ...

    def compute_inverse_footprint(  # type: ignore
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Compute the inverse footprint of the `foot` array, e.g. for expanding
        pointwise check fails into the points that could have contributed to
        the failures.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the inverse footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the inverse footprint at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The inverse footprint of the `foot` array.
        """

        ...

    @override
    def get_config(self) -> dict[str, JSON]:  # type: ignore
        """
        Returns the configuration of the safeguard.

        Returns
        -------
        config : dict[str, JSON]
            Configuration of the safeguard.
        """

        ...


class _AllSafeguardsBase(ABC):
    __slots__: tuple[str, ...] = ()

    kind: ClassVar[str] = "all"

    @property
    def safeguards(self) -> tuple[PointwiseSafeguard | StencilSafeguard, ...]:
        return self._safeguards  # type: ignore

    @property
    def late_bound(self) -> Set[Parameter]:
        return frozenset(b for s in self.safeguards for b in s.late_bound)

    def check_pointwise(
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        front, *tail = self.safeguards

        # pointwise check succeeds if all safeguards' pointwise check succeeds
        ok = front.check_pointwise(data, prediction, late_bound=late_bound, where=where)

        for safeguard in tail:
            ok &= safeguard.check_pointwise(
                data, prediction, late_bound=late_bound, where=where
            )

        return ok

    def compute_safe_intervals(
        self,
        data: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> IntervalUnion[T, int, int]:
        front, *tail = self.safeguards

        # only the intersection of all safeguards' safe intervals is safe
        valid = front.compute_safe_intervals(data, late_bound=late_bound, where=where)

        for safeguard in tail:
            valid = valid.intersect(
                safeguard.compute_safe_intervals(
                    data, late_bound=late_bound, where=where
                )
            )

        return valid

    def compute_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        front, *tail = self.safeguards

        # since all safeguards can contribute to the footprint, the union of
        #  their footprints forms the combined footprint
        footprint = front.compute_footprint(foot, late_bound=late_bound, where=where)

        for safeguard in tail:
            footprint |= safeguard.compute_footprint(
                foot, late_bound=late_bound, where=where
            )

        return footprint

    def compute_inverse_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        front, *tail = self.safeguards

        # since all safeguards can contribute to the footprint, the union of
        #  their footprints forms the combined footprint
        footprint = front.compute_inverse_footprint(
            foot, late_bound=late_bound, where=where
        )

        for safeguard in tail:
            footprint |= safeguard.compute_inverse_footprint(
                foot, late_bound=late_bound, where=where
            )

        return footprint

    def get_config(self) -> dict[str, JSON]:
        return dict(
            kind=type(self).kind,
            safeguards=[safeguard.get_config() for safeguard in self.safeguards],
        )

    @override
    def __repr__(self) -> str:
        return f"{AllSafeguards.__name__}(safeguards={list(self.safeguards)!r})"


class _AllPointwiseSafeguards(_AllSafeguardsBase, PointwiseSafeguard):
    __slots__: tuple[str, ...] = ("_safeguards",)
    _safeguards: tuple[PointwiseSafeguard, ...]

    def __init__(self, *safeguards: PointwiseSafeguard) -> None:
        for safeguard in safeguards:
            assert isinstance(safeguard, PointwiseSafeguard), (
                f"{safeguard!r} is not a pointwise safeguard"
            )
        self._safeguards = safeguards


class _AllStencilSafeguards(_AllSafeguardsBase, StencilSafeguard):
    __slots__: tuple[str, ...] = ("_safeguards",)
    _safeguards: tuple[PointwiseSafeguard | StencilSafeguard, ...]

    def __init__(self, *safeguards: PointwiseSafeguard | StencilSafeguard) -> None:
        for safeguard in safeguards:
            assert isinstance(safeguard, PointwiseSafeguard | StencilSafeguard), (
                f"{safeguard!r} is not a pointwise or stencil safeguard"
            )
        self._safeguards = safeguards

    @override
    def compute_check_neighbourhood_for_data_shape(
        self, data_shape: tuple[int, ...]
    ) -> tuple[dict[BoundaryCondition, NeighbourhoodAxis], ...]:
        neighbourhood: list[dict[BoundaryCondition, NeighbourhoodAxis]] = [
            dict() for _ in data_shape
        ]

        for safeguard in self._safeguards:
            if not isinstance(safeguard, StencilSafeguard):
                continue

            safeguard_neighbourhood = (
                safeguard.compute_check_neighbourhood_for_data_shape(data_shape)
            )

            for i, sn in enumerate(safeguard_neighbourhood):
                ni = neighbourhood[i]

                for b, s in sn.items():
                    if b in ni:
                        neighbourhood[i][b] = NeighbourhoodAxis(
                            max(ni[b].before, s.before), max(ni[b].after, s.after)
                        )
                    else:
                        neighbourhood[i][b] = s

        return tuple(neighbourhood)



"""
Always safe (logical truth) combinator safeguard.
"""

__all__ = ["AssumeAlwaysSafeguard"]

from collections.abc import Set
from typing import ClassVar, Literal

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ...utils._compat import _ones
from ...utils.bindings import Bindings, Parameter
from ...utils.intervals import Interval, IntervalUnion
from ...utils.typing import JSON, S, T
from ..pointwise.abc import PointwiseSafeguard


class AssumeAlwaysSafeguard(PointwiseSafeguard):
    """
    The `AssumeAlwaysSafeguard` assumes that all elements always meet their guarantees
    and are thus always safe.

    This safeguards can be used with the
    [`SelectSafeguard`][...select.SelectSafeguard] to express regions that are
    *not* of interest, i.e. where no additional safety requirements are imposed.
    """

    __slots__: tuple[str, ...] = ()

    kind: ClassVar[str] = "assume_safe"

    def __init__(self) -> None:
        pass

    @property
    @override
    def late_bound(self) -> Set[Parameter]:
        """
        The set of late-bound parameters that this safeguard has.

        Late-bound parameters are only bound when checking and applying the
        safeguard, in contrast to the normal early-bound parameters that are
        configured during safeguard initialisation.

        Late-bound parameters can be used for parameters that depend on the
        specific data that is to be safeguarded.
        """

        return frozenset()

    @override
    def check_pointwise(
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        All elements are safe and thus always succeed the check.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Original data array, relative to which the `prediction` is checked.
        prediction : np.ndarray[S, np.dtype[T]]
            Prediction for the `data` array.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only check at data points where the condition is [`True`][True].

        Returns
        -------
        ok : np.ndarray[S, np.dtype[np.bool]]
            Pointwise, `True` for every element.
        """

        return _ones(data.shape, dtype=np.dtype(np.bool))

    @override
    def compute_safe_intervals(
        self,
        data: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> IntervalUnion[T, int, int]:
        """
        Since all values are always safe, the safe intervals for each element
        in the `data` array are full (contain all possible values).

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Data for which the safe intervals should be computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the safe intervals at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        intervals : IntervalUnion[T, int, int]
            Union of intervals that cover all possible values, since all are
            always safe.
        """

        return Interval.full_like(data).into_union()

    @override
    def compute_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Since all values are always safe, the footprint of the `foot` array is
        [`False`][False] everywhere.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the footprint at pointwise checks where the condition
            is [`True`][True].

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The footprint of the `foot` array.
        """

        return np.zeros_like(foot)

    @override
    def compute_inverse_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Since all values are always safe, the inverse footprint of the `foot`
        array is [`False`][False] everywhere.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the inverse footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the inverse footprint at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The inverse footprint of the `foot` array.
        """

        return np.zeros_like(foot)

    @override
    def get_config(self) -> dict[str, JSON]:
        """
        Returns the configuration of the safeguard.

        Returns
        -------
        config : dict
            Configuration of the safeguard.
        """

        return dict(kind=type(self).kind)



from collections.abc import Callable, Mapping
from enum import Enum, auto

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import (
    _ensure_array,
    _is_negative_zero,
    _maximum_zero_sign_sensitive,
    _minimum_zero_sign_sensitive,
)
from ....utils.bindings import Parameter
from ..bound import checked_data_bounds, guarantee_arg_within_expr_bounds
from ..typing import F, Ns, Ps, np_sndarray
from .abc import AnyExpr, Expr
from .constfold import ScalarFoldedConstant


class Logarithm(Enum):
    ln = auto()
    log2 = auto()
    log10 = auto()


class ScalarLog(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_log", "_a")
    _log: Logarithm
    _a: AnyExpr

    def __init__(self, log: Logarithm, a: AnyExpr) -> None:
        self._log = log
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarLog":
        return ScalarLog(self._log, a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a,
            dtype,
            LOGARITHM_UFUNC[self._log],  # type: ignore
            lambda e: ScalarLog(self._log, e),
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return (LOGARITHM_UFUNC[self._log])(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg and log(arg)
        arg = self._a
        argv = arg.eval(Xs, late_bound)
        exprv = (LOGARITHM_UFUNC[self._log])(argv)

        smallest_subnormal = np.finfo(Xs.dtype).smallest_subnormal

        # apply the inverse function to get the bounds on arg
        # log(...) is NaN for negative values and can then take any negative
        #  value
        # if arg_lower == argv and argv == -0.0, we need to guarantee that
        #  arg_lower is also -0.0, same for arg_upper
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            _minimum_zero_sign_sensitive(
                argv, (LOGARITHM_EXPONENTIAL_UFUNC[self._log])(expr_lower)
            )
        )
        arg_lower[np.less(argv, 0)] = -np.inf

        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            _maximum_zero_sign_sensitive(
                argv, (LOGARITHM_EXPONENTIAL_UFUNC[self._log])(expr_upper)
            )
        )
        arg_upper[np.less(argv, 0)] = -smallest_subnormal

        # an upper bound of -0.0 must be nudged downwards from 1 to 1-eps
        arg_upper[_is_negative_zero(expr_upper)] = np.nextafter(
            np.array(Xs.dtype.type(1)), np.array(Xs.dtype.type(0))
        )

        # we need to force argv if expr_lower == expr_upper
        np.copyto(arg_lower, argv, where=(expr_lower == expr_upper), casting="no")
        np.copyto(arg_upper, argv, where=(expr_lower == expr_upper), casting="no")

        # handle rounding errors in log(exp(...)) early
        arg_lower = guarantee_arg_within_expr_bounds(
            lambda arg_lower: (LOGARITHM_UFUNC[self._log])(arg_lower),
            exprv,
            argv,
            arg_lower,
            expr_lower,
            expr_upper,
        )
        arg_upper = guarantee_arg_within_expr_bounds(
            lambda arg_upper: (LOGARITHM_UFUNC[self._log])(arg_upper),
            exprv,
            argv,
            arg_upper,
            expr_lower,
            expr_upper,
        )

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"{self._log.name}({self._a!r})"


LOGARITHM_UFUNC: dict[
    Logarithm,
    Callable[[np.ndarray[tuple[Ps], np.dtype[F]]], np.ndarray[tuple[Ps], np.dtype[F]]],
] = {
    Logarithm.ln: np.log,
    Logarithm.log2: np.log2,
    Logarithm.log10: np.log10,
}


class Exponential(Enum):
    exp = auto()
    exp2 = auto()
    exp10 = auto()


class ScalarExp(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_exp", "_a")
    _exp: Exponential
    _a: AnyExpr

    def __init__(self, exp: Exponential, a: AnyExpr) -> None:
        self._exp = exp
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarExp":
        return ScalarExp(self._exp, a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a,
            dtype,
            EXPONENTIAL_UFUNC[self._exp],  # type: ignore
            lambda e: ScalarExp(self._exp, e),
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return (EXPONENTIAL_UFUNC[self._exp])(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg and exp(arg)
        arg = self._a
        argv = arg.eval(Xs, late_bound)
        exprv = (EXPONENTIAL_UFUNC[self._exp])(argv)

        # apply the inverse function to get the bounds on arg
        # exp(...) cannot be negative, so ensure the bounds on expr also cannot
        # if arg_lower == argv and argv == -0.0, we need to guarantee that
        #  arg_lower is also -0.0, same for arg_upper
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            _minimum_zero_sign_sensitive(
                argv,
                (EXPONENTIAL_LOGARITHM_UFUNC[self._exp])(
                    _maximum_zero_sign_sensitive(Xs.dtype.type(0), expr_lower)
                ),
            )
        )
        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            _maximum_zero_sign_sensitive(
                argv, (EXPONENTIAL_LOGARITHM_UFUNC[self._exp])(expr_upper)
            )
        )

        # we need to force argv if expr_lower == expr_upper
        np.copyto(arg_lower, argv, where=(expr_lower == expr_upper), casting="no")
        np.copyto(arg_upper, argv, where=(expr_lower == expr_upper), casting="no")

        # handle rounding errors in exp(log(...)) early
        arg_lower = guarantee_arg_within_expr_bounds(
            lambda arg_lower: (EXPONENTIAL_UFUNC[self._exp])(arg_lower),
            exprv,
            argv,
            arg_lower,
            expr_lower,
            expr_upper,
        )
        arg_upper = guarantee_arg_within_expr_bounds(
            lambda arg_upper: (EXPONENTIAL_UFUNC[self._exp])(arg_upper),
            exprv,
            argv,
            arg_upper,
            expr_lower,
            expr_upper,
        )

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"{self._exp.name}({self._a!r})"


EXPONENTIAL_UFUNC: dict[
    Exponential,
    Callable[[np.ndarray[tuple[Ps], np.dtype[F]]], np.ndarray[tuple[Ps], np.dtype[F]]],
] = {
    Exponential.exp: np.exp,
    Exponential.exp2: np.exp2,
    Exponential.exp10: lambda x: np.power(10, x),
}

LOGARITHM_EXPONENTIAL_UFUNC: dict[
    Logarithm,
    Callable[[np.ndarray[tuple[Ps], np.dtype[F]]], np.ndarray[tuple[Ps], np.dtype[F]]],
] = {
    Logarithm.ln: EXPONENTIAL_UFUNC[Exponential.exp],
    Logarithm.log2: EXPONENTIAL_UFUNC[Exponential.exp2],
    Logarithm.log10: EXPONENTIAL_UFUNC[Exponential.exp10],
}

EXPONENTIAL_LOGARITHM_UFUNC: dict[
    Exponential,
    Callable[[np.ndarray[tuple[Ps], np.dtype[F]]], np.ndarray[tuple[Ps], np.dtype[F]]],
] = {
    Exponential.exp: LOGARITHM_UFUNC[Logarithm.ln],
    Exponential.exp2: LOGARITHM_UFUNC[Logarithm.log2],
    Exponential.exp10: LOGARITHM_UFUNC[Logarithm.log10],
}


class ScalarLogWithBase(Expr[AnyExpr, AnyExpr]):
    __slots__: tuple[str, ...] = ("_a", "_b")
    _a: AnyExpr
    _b: AnyExpr

    def __init__(self, a: AnyExpr, b: AnyExpr) -> None:
        self._a = a
        self._b = b

    @property
    @override
    def args(self) -> tuple[AnyExpr, AnyExpr]:
        return (self._a, self._b)

    @override
    def with_args(self, a: AnyExpr, b: AnyExpr) -> "ScalarLogWithBase":
        return ScalarLogWithBase(a, b)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        from .divmul import ScalarDivide  # noqa: PLC0415

        return ScalarDivide(
            ScalarLog(
                Logarithm.ln,
                self._a,
            ),
            ScalarLog(
                Logarithm.ln,
                self._b,
            ),
        ).constant_fold(dtype)

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        from .divmul import ScalarDivide  # noqa: PLC0415

        return ScalarDivide(
            ScalarLog(
                Logarithm.ln,
                self._a,
            ),
            ScalarLog(
                Logarithm.ln,
                self._b,
            ),
        ).eval(Xs, late_bound)

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        from .divmul import ScalarDivide  # noqa: PLC0415

        return ScalarDivide(
            ScalarLog(
                Logarithm.ln,
                self._a,
            ),
            ScalarLog(
                Logarithm.ln,
                self._b,
            ),
        ).compute_data_bounds(expr_lower, expr_upper, Xs, late_bound)

    @override
    def __repr__(self) -> str:
        return f"log({self._a!r}, base={self._b!r})"



import operator
from collections.abc import Mapping

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils.bindings import Parameter
from ..bound import checked_data_bounds
from ..typing import F, Ns, Ps, np_sndarray
from .abc import AnyExpr, Expr
from .constfold import ScalarFoldedConstant
from .literal import Number


class ScalarNegate(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr) -> None:
        self._a = a

    def __new__(cls, a: AnyExpr) -> "ScalarNegate | Number":  # type: ignore[misc]
        na = Number.symbolic_fold_unary(a, operator.neg)
        if na is not None:
            return na
        return super().__new__(cls)

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarNegate | Number":
        return ScalarNegate(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a, dtype, np.negative, ScalarNegate
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.negative(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        return self._a.compute_data_bounds(-expr_upper, -expr_lower, Xs, late_bound)

    @override
    def __repr__(self) -> str:
        return f"-{self._a!r}"
